// -----------------------------------------------------------------------------
//
// Copyright (C) 2021 CERN & Newcastle University for the benefit of the
// BioDynaMo collaboration. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
//
// See the LICENSE file distributed with this work for details.
// See the NOTICE file distributed with this work for additional information
// regarding copyright ownership.
//
// -----------------------------------------------------------------------------
#ifndef RILIMODEL_H_
#define RILIMODEL_H_

#include <TTreeReader.h>
#include <TTreeReaderArray.h>
#include <TTreeReaderValue.h>
#include <math.h>
#include <time.h>
#include <chrono>
#include <ctime>
#include <vector>
#include "TROOT.h"
#include "TTree.h"
#include "alveolarDuct.h"
#include "biodynamo.h"
#include "core/operation/reduction_op.h"
#include "core/simulation_backup.h"
#include "core/substance_initializers.h"
#include "myParam.h"
#include "output_utils.h"

namespace bdm {

inline void importCellPositions(Simulation* sim, ResourceManager* rm,
                                double const celldiam) {
  // Import cell positions from ROOT file
  double x_new = 0, y_new = 0, z_new;

  const char* file_name =
      sim->GetParam()->Get<SimParam>()->input_file_name.c_str();

  TFile* f = new TFile(file_name);
  TTree* cellsTree = (TTree*)f->Get("tree");

  cellsTree->SetBranchAddress("x", &x_new);
  cellsTree->SetBranchAddress("y", &y_new);
  cellsTree->SetBranchAddress("z", &z_new);

  int cellsEntries = cellsTree->GetEntries();

  for (int i = 0; i < cellsEntries; ++i) {
    cellsTree->GetEntry(i);
    auto* cell = new alveolarDuct();
    cell->SetPosition({x_new, y_new, z_new});
    cell->SetDiameter(celldiam);
    rm->AddAgent(cell);
  }
}

// Use this function to generate the geometrical structure
inline void dichotomicTreeAndExportCellPositions(ResourceManager* rm,
                                                 double const celldiam) {
  // Generate dichotomic tree
  segment Segments[nb_of_segments];
  segmentStructure(Segments);

  // Get cell positions from segment array
  for (size_t seg = 0; seg < nb_of_segments; seg++) {
    auto* cell = new alveolarDuct();
    cell->SetPosition(Segments[seg].center);
    cell->SetDiameter(celldiam);

    rm->AddAgent(cell);
  }

  // Export cell positions
  std::vector<Double3> positions;

  for (size_t seg = 0; seg < nb_of_segments; seg++) {
    positions.push_back(Segments[seg].center);
  }

  Double_t x, y, z;

  TFile* file = new TFile("CellsData.root", "RECREATE");
  TTree* tree = new TTree("tree", "Testing root trees");
  tree->Branch("x", &x, "x/D");
  tree->Branch("y", &y, "y/D");
  tree->Branch("z", &z, "z/D");

  for (unsigned long int event = 0; event < positions.size(); event++) {
    x = positions[event][0];
    y = positions[event][1];
    z = positions[event][2];

    tree->Fill();
  }

  // Write/Export tree
  tree->Print();
  tree->Write();
  file->Close();
}

// Utility for const conc substances initialization
struct FlatConc {
  double initial_conc_;

  FlatConc(double initial_conc) { initial_conc_ = initial_conc; }

  double operator()(double x, double y, double z) { return initial_conc_; }
};

inline void define_init_ExtracellularSubstances(Simulation* sim,
                                                int resolution) {
  auto* simPar = sim->GetParam()->Get<SimParam>();

  ModelInitializer::DefineSubstance(
      kPDGF, "PDGF", simPar->PDGF_D / (double)seconds_per_day_,
      simPar->PDGF_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kPDGF,
                                        FlatConc(simPar->PDGF_flat_conc));

  ModelInitializer::DefineSubstance(
      kMMP, "MMP", simPar->MMP_D / (double)seconds_per_day_,
      simPar->MMP_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0,
      simPar->TIMP_MMP_binding / (double)seconds_per_day_, kTIMP);

  ModelInitializer::InitializeSubstance(kMMP, FlatConc(simPar->MMP_flat_conc));

  ModelInitializer::DefineSubstance(
      kECM, "ECM", simPar->ECM_D / (double)seconds_per_day_,
      simPar->ECM_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0,
      simPar->MMP_ECM_binding / (double)seconds_per_day_, kMMP);

  ModelInitializer::InitializeSubstance(kECM, FlatConc(simPar->ECM_flat_conc));

  ModelInitializer::DefineSubstance(
      kMCP1, "MCP-1", simPar->MCP1_D / (double)seconds_per_day_,
      simPar->MCP1_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kMCP1,
                                        FlatConc(simPar->MCP1_flat_conc));

  ModelInitializer::DefineSubstance(
      kTIMP, "TIMP", simPar->TIMP_D / (double)seconds_per_day_,
      simPar->TIMP_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0,
      simPar->MMP_TIMP_binding / (double)seconds_per_day_, kMMP);

  ModelInitializer::InitializeSubstance(kTIMP,
                                        FlatConc(simPar->TIMP_flat_conc));

  ModelInitializer::DefineSubstance(
      kTGFb_in, "TGFb-inactive", simPar->TGFb_in_D / (double)seconds_per_day_,
      simPar->TGFb_in_d / (double)seconds_per_day_, resolution, {}, 0, {},
      0);  // s

  ModelInitializer::InitializeSubstance(kTGFb_in,
                                        FlatConc(simPar->TGFb_in_flat_conc));

  ModelInitializer::DefineSubstance(
      kTGFb_ac, "TGFb-active", simPar->TGFb_ac_D / (double)seconds_per_day_,
      simPar->TGFb_ac_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kTGFb_ac,
                                        FlatConc(simPar->TGFb_ac_flat_conc));

  ModelInitializer::DefineSubstance(
      kTNFa, "TNFa", simPar->TNFa_D / (double)seconds_per_day_,
      simPar->TNFa_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kTNFa,
                                        FlatConc(simPar->TNFa_flat_conc));

  ModelInitializer::DefineSubstance(
      kIL13, "IL-13", simPar->IL13_D / (double)seconds_per_day_,
      simPar->IL13_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kIL13,
                                        FlatConc(simPar->IL13_flat_conc));

  ModelInitializer::DefineSubstance(
      kFGF2, "FGF-2", simPar->FGF2_D / (double)seconds_per_day_,
      simPar->FGF2_d / (double)seconds_per_day_, resolution, {}, 0, {}, 0);

  ModelInitializer::InitializeSubstance(kFGF2,
                                        FlatConc(simPar->FGF2_flat_conc));
}

inline void setBehaviors(Simulation* sim) {
  // Add behaviours to cells
  sim->GetResourceManager()->ForEachAgent([](Agent* element) {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(element)) {
      // Secretion/Degradation behaviors
      cell->AddBehavior(new AEC2_TNFaSecretion());
      cell->AddBehavior(new AEC2_MCP1Secretion());
      cell->AddBehavior(new AEC2_FGF2Secretion());
      cell->AddBehavior(new AEC2_TGFbSecretion());
      cell->AddBehavior(new F_TGFbSecretion());
      cell->AddBehavior(new F_ECMSecretion());
      cell->AddBehavior(new M2_IL13Secretion());
      cell->AddBehavior(new M2_PDGFSecretion());
      cell->AddBehavior(new M2_MMPSecretion());
      cell->AddBehavior(new M2_TIMPSecretion());
      cell->AddBehavior(new M2_TGFbSecretion());
      cell->AddBehavior(new M1_TNFaSecretion());
      cell->AddBehavior(new L_IL13Secretion());
      cell->AddBehavior(new MF_ECMSecretion());
      // Proliferation behaviors
      cell->AddBehavior(new F_Proliferate());
      cell->AddBehavior(new AEC2_Proliferate());
      // Apoptosis behaviors
      cell->AddBehavior(new Apoptosis());
      // Damage/repair behaviors
      cell->AddBehavior(new AEC2_Activation());
      cell->AddBehavior(new AEC2_Repair());
      // Differentiation behaviors
      cell->AddBehavior(new F_MF_Differentiation());
      cell->AddBehavior(new M12_Differentiation());
      cell->AddBehavior(new M21_Differentiation());
      cell->AddBehavior(new MO1_Differentiation());
      cell->AddBehavior(new AEC21_Differentiation());
      // cell->AddBehavior(new AEC2_EMT());
      // Add cell behaviors
      cell->AddBehavior(new F_Addition());
    }
  });
}

// Functor for reduction op (count cell of each type)
struct CheckCellType : public Functor<void, Agent*, double*> {
  explicit CheckCellType(int t) : type_(t) {}

  void operator()(Agent* so, double* tl_result) {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      (*tl_result) += cell->getCurrentCellNumber(type_);
    }
  }

  int type_;
};

// Functor for reduction op (get conc from each cell)
struct GetCellConc : public Functor<void, Agent*, double*> {
  explicit GetCellConc(size_t s) : sub_(s) {}

  void operator()(Agent* so, double* tl_result) {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      (*tl_result) += cell->getLocalConc(sub_);
    }
  }

  size_t sub_;
};

inline void setupAndScheduleOps(
    Simulation* sim, std::vector<ConcData*>* concentrations,
    bdm::ReductionOp<double>* Count_cellType_op_impl[], size_t nb_of_cell_types,
    bdm::ReductionOp<double>* CellsMeasureConc_op_impl[], size_t nb_of_subs, std::vector<std::vector<double>>* fullECMmap_) {
  auto* simPar = sim->GetParam()->Get<SimParam>();

  // Initialize the concentration vector with the pointers
  for (size_t i = 0; i < nb_of_subs; i++) {
    concentrations->push_back(new bdm::ConcData());
  }

  // We instantiate the operations
  // Conc. measurement ops
  Operation* MeasureConcs[nb_of_subs];      // Global concs
  Operation* CellsMeasureConc[nb_of_subs];  // Local conc in cell positions

  // Cells count op
  Operation* Count_cellTypes_[nb_of_cell_types];  // Count total number of
  // simulation objects with a
  // certain cell type

  // Cells count op
  for (size_t i = 0; i < nb_of_cell_types; i++) {
    Count_cellTypes_[i] = NewOperation("ReductionOpDouble");
    Count_cellType_op_impl[i] =
        Count_cellTypes_[i]->GetImplementation<ReductionOp<double>>();
    Count_cellType_op_impl[i]->Initialize(new CheckCellType(i + 1),
                                          new SumReduction<double>());
    Count_cellTypes_[i]->frequency_ = simPar->measurement_cells_ops_freq_;
  }

  // Change the substance values of each instantiation
  // We have to specify the type of the operation implementation
  // 'MeasureConcOp'
  for (size_t i = 0; i < nb_of_subs; i++) {

    // Conc from cells
    CellsMeasureConc[i] = NewOperation("ReductionOpDouble");
    CellsMeasureConc_op_impl[i] =
        CellsMeasureConc[i]->GetImplementation<ReductionOp<double>>();
    CellsMeasureConc_op_impl[i]->Initialize(new GetCellConc(i),
                                            new SumReduction<double>());
    CellsMeasureConc[i]->frequency_ = simPar->measurement_conc_ops_freq_;
  }

  // Cells count op
  for (size_t i = 0; i < nb_of_cell_types; i++) {
    sim->GetScheduler()->ScheduleOp(Count_cellTypes_[i]);
  }

  // Schedule all operations
  for (size_t i = 0; i < nb_of_subs; i++) {
    sim->GetScheduler()->ScheduleOp(
        CellsMeasureConc[i]);  // Get local average conc from cell positions
  }
}

inline void exportCellsAndConcentrations(
    Simulation* sim, std::vector<ConcData*>* concentrations,
    bdm::ReductionOp<double>* Count_cellType_op_impl[], size_t nb_of_cell_types,
    bdm::ReductionOp<double>* CellsMeasureConc_op_impl[], size_t nb_of_subs, std::vector<std::vector<double>>* fullECMmap_) {
  auto* simPar = sim->GetParam()->Get<SimParam>();
  auto* random = sim->GetRandom();
  size_t nb_of_sim_steps = simPar->nb_of_sim_steps_;

  // Make output directory
  std::time_t now =
      std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
  std::string s(16, '\0');
  std::strftime(&s[0], s.size(), "%d%m%Y_%H%M%S", std::localtime(&now));
  std::string outputDir =
      Concat("results/", simPar->nb_of_sim_steps_, "_", s).c_str();
  if (simPar->setInitialDamageAEC2 == true) {
    outputDir = Concat(outputDir, "_dam").c_str();
  }
  outputDir = Concat(outputDir, "_", (int)random->Uniform(0, 1000))
                  .c_str();  // Add a random number to avoid directory overwrite
                             // when running multiple simulations
  system(Concat("mkdir -p ", outputDir).c_str());

  // Cells export
  std::vector<std::vector<double>> cell_counts_ops_results;
  for (size_t i = 0; i < nb_of_cell_types; i++) {
    cell_counts_ops_results.push_back(Count_cellType_op_impl[i]->GetResults());
  }

  // Local conc export
  std::vector<std::vector<double>> local_conc_ops_results;
  for (size_t i = 0; i < nb_of_subs; i++) {
    local_conc_ops_results.push_back(CellsMeasureConc_op_impl[i]->GetResults());
  }

  size_t nb_of_agents = sim->GetResourceManager()->GetNumAgents();

  for (size_t i = 0; i < local_conc_ops_results.size(); i++) {
    for (size_t j = 0; j < local_conc_ops_results.at(i).size(); j++) {
      local_conc_ops_results.at(i).at(j) /=
          nb_of_agents;  // Divide by the nb of measurements to compute the avg
    }
  }

  export_to_root_cells("Cells vs time", cell_counts_ops_results,
                       nb_of_sim_steps, nb_of_cell_types,
                       simPar->measurement_cells_ops_freq_, outputDir,
                       nb_of_alveoli);

  export_to_root_local_concs("Local conc vs time", local_conc_ops_results,
                             nb_of_sim_steps, nb_of_subs,
                             simPar->measurement_conc_ops_freq_, outputDir);

  // Copy parameters to outpur directory (use the same code used in BDM to
  // export statistics)
  std::stringstream sstr;
  sstr << *sim << std::endl;
  // write to file
  std::ofstream ofs(Concat(outputDir, "/param.txt"));
  ofs << sstr.str() << std::endl;

  // export ECM map (only at the last time step)
  auto* rm = sim->GetResourceManager();
  static auto* ecmDg = rm->GetDiffusionGrid(kECM);
  std::vector<double> ecmConcMap;  // Get conc distribution at last timestep

  sim->GetResourceManager()->ForEachAgent(
      [&ecmConcMap](Agent* so) {  // Each agent is a duct
        if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
          ecmConcMap.push_back(ecmDg->GetConcentration(cell->GetPosition()));
        }
      });

  export_to_root_ecm_map("ECM conc map final", ecmConcMap, outputDir);

}

inline void addInitialDamage(Simulation* sim) {
  Log::Warning("Added initial damage");

  // Uniform damage
  if (sim->GetParam()->Get<SimParam>()->damageType == 0) {
    sim->GetResourceManager()->ForEachAgent(
        [](Agent* so) {  // Each agent is a duct
          if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
            auto* sim = Simulation::GetActive();
            int old_type_ = 2, new_type_ = 13;
            auto* sparam = sim->GetParam()->Get<SimParam>();

            double hitProb = sparam->hitProbability;

            double current_AEC2 = cell->getCurrentCellNumber(old_type_);
            double current_AEC2_damaged = cell->getCurrentCellNumber(new_type_);

            cell->setCurrentCellNumber(old_type_, current_AEC2 * (1 - hitProb));
            cell->setCurrentCellNumber(
                new_type_, current_AEC2_damaged + current_AEC2 * hitProb);
          }
        });
  }

  // Heterogeneous damage
  if (sim->GetParam()->Get<SimParam>()->damageType == 2) {
    sim->GetResourceManager()->ForEachAgent(
        [](Agent* so) {  // Each agent is a duct
          if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
            auto* sim = Simulation::GetActive();
            int old_type_ = 2, new_type_ = 13;
            auto* sparam = sim->GetParam()->Get<SimParam>();
            auto* rnd = sim->GetRandom();

            double hitProb = rnd->Gaus(sparam->hitProbability, 0.01);
            if (hitProb< 0.0)
            {
              hitProb = 0.0;
            }
            if (hitProb> 100.0)
            {
              hitProb = 100.0;
            }

            Log::Warning("hitProb", hitProb);

            double current_AEC2 = cell->getCurrentCellNumber(old_type_);
            double current_AEC2_damaged = cell->getCurrentCellNumber(new_type_);

            cell->setCurrentCellNumber(old_type_, current_AEC2 * (1 - hitProb));
            cell->setCurrentCellNumber(
                new_type_, current_AEC2_damaged + current_AEC2 * hitProb);
          }
        });
  }
}

// Main function
inline int Simulate(int argc, const char** argv) {
  Param::RegisterParamGroup(new SimParam());
  Simulation simulation(argc, argv, {"bdm.json"});
  auto* random = simulation.GetRandom();
  auto* simPar = Simulation::GetActive()->GetParam()->Get<SimParam>();
  auto* rm = simulation.GetResourceManager();
  random->SetSeed(
      0);  // Set seed to get different random values for each new simulation

  int resolution = simPar->sub_res;
  // dichotomicTreeAndExportCellPositions(rm, cell_diameter); // Use this function to generate the geometrical structure
  double const cell_diameter = 700;
  importCellPositions(&simulation, rm, cell_diameter);
  setBehaviors(&simulation);
  std::vector<ConcData*>* concentrations =
      new std::vector<ConcData*>;  // Vector of pointers to struct (which
                                   // contains vectors) to be given to the ops
  size_t nb_of_cell_types = simPar->nb_of_cell_types_;
  size_t nb_of_subs = simPar->nb_of_subs_;

  bdm::ReductionOp<double>* Count_cellType_op_impl[nb_of_cell_types];
  bdm::ReductionOp<double>*
      CellsMeasureConc_op_impl[nb_of_subs];  // Local conc from cells
  setupAndScheduleOps(&simulation, concentrations, Count_cellType_op_impl,
                      nb_of_cell_types, CellsMeasureConc_op_impl, nb_of_subs);

  if (simPar->setInitialDamageAEC2 == true) {  // Damage AEC2 cells
    addInitialDamage(&simulation);
  }

  define_init_ExtracellularSubstances(&simulation, resolution);

  // Run simulation for one timestep
  simulation.GetScheduler()->Simulate(simPar->nb_of_sim_steps_);

  exportCellsAndConcentrations(&simulation, concentrations,
                               Count_cellType_op_impl, nb_of_cell_types,
                               CellsMeasureConc_op_impl, nb_of_subs);

  std::cout << "Simulation completed successfully!" << std::endl;

  // Cleanings
  for (auto* conc_i : *concentrations) {  // Clean the vec pointers
    delete conc_i;
  }

  return 0;
}

}  // namespace bdm

#endif  // RILIMODEL_H_
