// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#ifndef ALVEOLAR_DUCT_H_
#define ALVEOLAR_DUCT_H_

#include "biodynamo.h"
#include "core/agent/cell.h"
#include "geometry.h"
#include "myParam.h"

namespace bdm {

long int const seconds_per_day_ = 864000;  // to match time_step = 0.1

class alveolarDuct : public Cell {  // our object extends the Cell object
  // create the header with our new data member
  BDM_AGENT_HEADER(alveolarDuct, Cell, 1);

 public:
  alveolarDuct() {
    // Copy initial cell nb into current cell nb
    std::copy(std::begin(init_cell_nb), std::end(init_cell_nb),
              std::begin(current_cell_nb));
    auto* sim = Simulation::GetActive();
    auto* simPar = sim->GetParam()->Get<SimParam>();

    // Ones which are left empty are = 0. due to the fact that apoptosis isn't
    // defined for all cell types
    apoptosis_rates[0] =
        simPar->AEC1_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[1] =
        simPar->AEC2_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[2] = 0.;
    apoptosis_rates[3] =
        simPar->F_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[4] =
        simPar->M2_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[5] = 0.;
    apoptosis_rates[6] =
        simPar->M1_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[7] = 0.;
    apoptosis_rates[8] = 0.;
    apoptosis_rates[9] = 0.;
    apoptosis_rates[10] =
        simPar->MF_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[11] = 0.;
    apoptosis_rates[12] = 0.;
    apoptosis_rates[13] =
        simPar->active_AEC2_apoptosis / (double)seconds_per_day_;  // s^(-1)
  }

  explicit alveolarDuct(const Double3& position) : Base(position) {
    std::copy(std::begin(init_cell_nb), std::end(init_cell_nb),
              std::begin(current_cell_nb));
    auto* sim = Simulation::GetActive();
    auto* simPar = sim->GetParam()->Get<SimParam>();
    // Ones which are left empty are = 0. due to the fact that apoptosis isn't
    // defined for all cell types
    apoptosis_rates[0] =
        simPar->AEC1_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[1] =
        simPar->AEC2_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[2] = 0.;
    apoptosis_rates[3] =
        simPar->F_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[4] =
        simPar->M2_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[5] = 0.;
    apoptosis_rates[6] =
        simPar->M1_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[7] = 0.;
    apoptosis_rates[8] = 0.;
    apoptosis_rates[9] = 0.;
    apoptosis_rates[10] =
        simPar->MF_apoptosis / (double)seconds_per_day_;  // s^(-1)
    apoptosis_rates[11] = 0.;
    apoptosis_rates[12] = 0.;
    apoptosis_rates[13] =
        simPar->active_AEC2_apoptosis / (double)seconds_per_day_;  // s^(-1)
  }

  virtual ~alveolarDuct() {}

  /// If MyCell divides, the daughter has to initialize its attributes
  void Initialize(const NewAgentEvent& event) override {
    Base::Initialize(event);
  }

  double const getInitialCellNumber(size_t t) {
    if (t > 0) {
      return init_cell_nb[t - 1];
    } else {
      Log::Fatal("getInitialCellNumber(), index must be bigger than 0");
      return 0.;
    }
  }

  double getCurrentCellNumber(size_t t) {
    if (t > 0) {
      return current_cell_nb[t - 1];
    } else {
      Log::Fatal("getCurrentCellNumber(), index must be bigger than 0");
      return 0.;
    }
  }

  void setCurrentCellNumber(size_t t, double n) {
    if (t > 0) {
      if (n >= 0.) {
        current_cell_nb[t - 1] = n;
      } else {
        current_cell_nb[t - 1] = 0.;
      }
    } else {
      Log::Fatal("setCurrentCellNumber(), index must be bigger than 0");
    }
  }

  void updateCurrentCellNumber(size_t t, double m) {
    if (t > 0) {
      if ((current_cell_nb[t - 1] * m) >= 0.) {
        current_cell_nb[t - 1] = current_cell_nb[t - 1] * m;
      } else {
        current_cell_nb[t - 1] = 0.;
      }
    } else {
      Log::Fatal("updateCurrentCellNumber(), index must be bigger than 0");
    }
  }

  void updateAllCurrentCellNumber(double m) {

    for (size_t t = 1; t <=11; t++)
    {
      if ((current_cell_nb[t - 1] * m) >= 0.) {
        current_cell_nb[t - 1] = current_cell_nb[t - 1] * m;
      } else {
        current_cell_nb[t - 1] = 0.;
      }
    }

    current_cell_nb[11] = current_cell_nb[6]*5; // for the monocytes

  }

  double const getCellDeathRate(size_t t) {
    if (t > 0) {
      return apoptosis_rates[t - 1];
    } else {
      Log::Fatal("getCellDeathRate(), index must be bigger than 0");
      return 0.;
    }
  }

  void updateCellDeathRate(size_t t, double m) {
    if (t > 0) {
      if ((apoptosis_rates[t - 1] * m) >= 0.) {
        apoptosis_rates[t - 1] *= m;
      } else {
        apoptosis_rates[t - 1] = 0.;
      }
    } else {
      Log::Fatal("updateCellDeathRate(), index must be bigger than 0");
    }
  }

  double getLocalConc(size_t sub) {
    auto* subDg = Simulation::GetActive()->GetResourceManager()->GetDiffusionGrid(sub);
    return subDg->GetConcentration(this->GetPosition());
  }

 private:
  // Initial cell numbers (per alveolus * alveoli per segment/duct)
  double const init_nb_of_aec1 = 41 * alveoli_per_segment;            // Type 1
  double const init_nb_of_aec2 = 69 * alveoli_per_segment;            // Type 2
  double const init_nb_of_endo_cells = 136 * alveoli_per_segment;     // Type 3 not used
  double const init_nb_of_fibroblasts = 24 * alveoli_per_segment;     // Type 4
  double const init_nb_of_alv_macr = 12 * alveoli_per_segment;        // Type 5
  double const init_nb_of_pericytes = 24 * alveoli_per_segment;       // Type 6 not used
  double const init_nb_of_inter_macr = 13 * alveoli_per_segment;      // Type 7
  double const init_nb_of_lymphocytes = 4 * alveoli_per_segment;      // Type 8
  double const init_nb_of_plasma_cells = 2 * alveoli_per_segment;     // Type 9 not used
  double const init_nb_of_mast_cells = 4 * alveoli_per_segment;       // Type 10 not used
  double const init_nb_of_myofibroblasts = 36 * alveoli_per_segment;  // Type 11
  double const init_nb_of_monocytes = init_nb_of_inter_macr * 5;       // Type 12
  double const init_nb_of_aec2_damaged = 0 * alveoli_per_segment;     // Type 13
  double const init_nb_of_aec2_activated = 0 * alveoli_per_segment;   // Type 14
  double const init_th_nb_of_M0_per_segment =
      init_nb_of_inter_macr * 5;  // Type 15
  double const init_nb_of_interstitial_cells =
      init_nb_of_fibroblasts + init_nb_of_pericytes + init_nb_of_inter_macr +
      init_nb_of_lymphocytes + init_nb_of_plasma_cells + init_nb_of_mast_cells +
      init_nb_of_myofibroblasts;
  double const init_nb_of_epi_cells = init_nb_of_aec1 + init_nb_of_aec2;
  double const init_cell_nb[17] = {init_nb_of_aec1,
                                   init_nb_of_aec2,
                                   init_nb_of_endo_cells,
                                   init_nb_of_fibroblasts,
                                   init_nb_of_alv_macr,
                                   init_nb_of_pericytes,
                                   init_nb_of_inter_macr,
                                   init_nb_of_lymphocytes,
                                   init_nb_of_plasma_cells,
                                   init_nb_of_mast_cells,
                                   init_nb_of_myofibroblasts,
                                   init_nb_of_monocytes,
                                   init_nb_of_aec2_damaged,
                                   init_nb_of_aec2_activated,
                                   init_th_nb_of_M0_per_segment,
                                   init_nb_of_interstitial_cells,
                                   init_nb_of_epi_cells};
  // Current cell numbers
  double current_cell_nb[17] = {0.0};
  // Apoptosis rates
  double apoptosis_rates[14] = {0.};
};

}  // namespace bdm

#endif  // CELLS_H_
