// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#ifndef BIOLOGY_MODULES_H_
#define BIOLOGY_MODULES_H_

#include "alveolarDuct.h"
#include "biodynamo.h"
#include "core/simulation.h"
#include "myParam.h"

namespace bdm {

// List the extracellular substances
enum Substances {
  kPDGF,    /*1*/
  kMMP,     /*2*/
  kECM,     /*3*/
  kMCP1,    /*4*/
  kTIMP,    /*5*/
  kTGFb_in, /*6*/
  kTGFb_ac, /*7*/
  kTNFa,    /*8*/
  kIL13,    /*9*/
  kFGF2     /*10*/
};

/////////////////////////////////////////////////////////////
//
// Define secretion/degradation behaviors:
//
/////////////////////////////////////////////////////////////

struct M2_PDGFSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M2_PDGFSecretion, Behavior, 1);

  M2_PDGFSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 5;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* pdgfDg = rm->GetDiffusionGrid(kPDGF);
      double amount =
          sparam->M2_PDGF_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (pdgfDg->GetBoxVolume() * um3_To_cm3);           // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      pdgfDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct M2_MMPSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M2_MMPSecretion, Behavior, 1);

  M2_MMPSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 5;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* mmpDg = rm->GetDiffusionGrid(kMMP);
      double amount =
          sparam->M2_MMP_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (mmpDg->GetBoxVolume() * um3_To_cm3);           // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      mmpDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct M2_TIMPSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M2_TIMPSecretion, Behavior, 1);

  M2_TIMPSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 5;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* timpDg = rm->GetDiffusionGrid(kTIMP);
      double amount =
          sparam->M2_TIMP_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (timpDg->GetBoxVolume() * um3_To_cm3);           // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      timpDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct M2_IL13Secretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M2_IL13Secretion, Behavior, 1);

  M2_IL13Secretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 5;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* il13Dg = rm->GetDiffusionGrid(kIL13);
      double amount =
          sparam->M2_IL13_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (il13Dg->GetBoxVolume() * um3_To_cm3);           // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      il13Dg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct M2_TGFbSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M2_TGFbSecretion, Behavior, 1);

  M2_TGFbSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 5;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      static auto* il13Dg =
          rm->GetDiffusionGrid(kIL13);  // M2 interact with IL-13
      double il13LocalCon = il13Dg->GetConcentration(cell->GetPosition());
      double amount =
          sparam->M2_TGFb_secretion *
          (1 + sparam->IL13_effect_M2_TGFb_secretion *
                   (il13LocalCon / (il13LocalCon + sparam->k_IL13))) /
          (double)seconds_per_day_;  // s^(-1)
                                     // amount*=2.2*pow(10,-11);//
      amount /= (tgfb_acDg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      tgfb_acDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct M1_TNFaSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(M1_TNFaSecretion, Behavior, 1);

  M1_TNFaSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 7;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* tnfaDg = rm->GetDiffusionGrid(kTNFa);
      double amount =
          sparam->M1_TNFa_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (tnfaDg->GetBoxVolume() * um3_To_cm3);           // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      tnfaDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct AEC2_TNFaSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_TNFaSecretion, Behavior, 1);

  AEC2_TNFaSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 14;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* tnfaDg = rm->GetDiffusionGrid(kTNFa);
      double amount =
          sparam->AEC2_TNFa_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (tnfaDg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type)/* /
       cell->getInitialCellNumber(cell_type)*/);  //Currently initial cell # is 0
      tnfaDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct AEC2_MCP1Secretion : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_MCP1Secretion, Behavior, 1);

  AEC2_MCP1Secretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 14;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* mcp1Dg = rm->GetDiffusionGrid(kMCP1);
      double amount =
          sparam->AEC2_MCP1_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (mcp1Dg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type)/* /
       cell->getInitialCellNumber(cell_type)*/);  //Currently initial cell # is 0
      mcp1Dg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct AEC2_TGFbSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_TGFbSecretion, Behavior, 1);

  AEC2_TGFbSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 14, AEC2_cell_type = 2;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      static auto* tgfb_inDg = rm->GetDiffusionGrid(kTGFb_in);
      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);

      // From Hao et al.: initial E0 conc ~= 0.8g/cm^3. KE=0.1g/cm^3 -> KE=1/8
      // E0

      double activeAEC = cell->getCurrentCellNumber(cell_type);
      double KE = (cell->getInitialCellNumber(AEC2_cell_type)) * 1. / 8.;

      double TGFb_inLocalConc =
          tgfb_inDg->GetConcentration(cell->GetPosition());

      double amount = TGFb_inLocalConc*(activeAEC / (activeAEC + KE));

      amount /= (double)seconds_per_day_;  // s^(-1)

        tgfb_inDg->ChangeConcentrationBy(cell->GetPosition(),
                                         (-amount));  // Reduce
        tgfb_acDg->ChangeConcentrationBy(cell->GetPosition(),
                                         amount);  // Increase (=transform)

    }
  }
};

struct AEC2_FGF2Secretion : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_FGF2Secretion, Behavior, 1);

  AEC2_FGF2Secretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 14;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      static auto* fgf2Dg = rm->GetDiffusionGrid(kFGF2);

      double FGF2productionRate =
          sparam->AEC2_FGF2_secretion / (double)seconds_per_day_;  // s^(-1)

      double k_TGFb_in_AEC2_FGF2 = sparam->k_TGFb_in_AEC2_FGF2;

      double tgfb_acLocalConc =
          tgfb_acDg->GetConcentration(cell->GetPosition());

      double amount =
          FGF2productionRate *
          (1 + (tgfb_acLocalConc / (tgfb_acLocalConc + k_TGFb_in_AEC2_FGF2)));

      amount /= (fgf2Dg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type)/* /
                 cell->getInitialCellNumber(cell_type)*/);  //Currently initial cell # is 0
      fgf2Dg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct F_TGFbSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(F_TGFbSecretion, Behavior, 1);

  F_TGFbSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 4;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* tgfb_inDg = rm->GetDiffusionGrid(kTGFb_in);
      double amount =
          sparam->F_TGFb_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (tgfb_inDg->GetBoxVolume() * um3_To_cm3);       // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      tgfb_inDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct F_ECMSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(F_ECMSecretion, Behavior, 1);

  F_ECMSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 4;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* ecmDg = rm->GetDiffusionGrid(kECM);

      double ECMLocalConc = ecmDg->GetConcentration(cell->GetPosition());

      double factor = 0.;
      double ECMsaturation = sparam->ECM_saturation;
      if (ECMLocalConc < ECMsaturation) {
        factor = (1.0 - (ECMLocalConc / ECMsaturation));
        double amount = (sparam->F_ECM_secretion) * factor /
                        (double)seconds_per_day_;        // s^(-1)
        amount /= (ecmDg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
        amount *= (cell->getCurrentCellNumber(cell_type) /
                   cell->getInitialCellNumber(cell_type));
        ecmDg->ChangeConcentrationBy(cell->GetPosition(), amount);
      }
    }
  }
};

struct MF_ECMSecretion : public Behavior {
  BDM_BEHAVIOR_HEADER(MF_ECMSecretion, Behavior, 1);

  MF_ECMSecretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 11;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();

      static auto* ecmDg = rm->GetDiffusionGrid(kECM);
      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      double TGFb_acLocalConc =
          tgfb_acDg->GetConcentration(cell->GetPosition());
      double amount =
          sparam->MF_ECM_secretion *
          (1 + sparam->TGFb_ac_effect_MF_ECM_secretion *
                   (TGFb_acLocalConc / (TGFb_acLocalConc + sparam->k_TGFb_ac)));
      amount /= (double)seconds_per_day_;              // s^(-1)
      amount /= (ecmDg->GetBoxVolume() * um3_To_cm3);  // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      ecmDg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

struct L_IL13Secretion : public Behavior {
  BDM_BEHAVIOR_HEADER(L_IL13Secretion, Behavior, 1);

  L_IL13Secretion() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 8;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      static auto* il13Dg = rm->GetDiffusionGrid(kIL13);
      double amount =
          sparam->L_IL13_secretion / (double)seconds_per_day_;  // s^(-1)
      amount /= (il13Dg->GetBoxVolume() * um3_To_cm3);          // conc. g/cm^3
      amount *= (cell->getCurrentCellNumber(cell_type) /
                 cell->getInitialCellNumber(cell_type));
      il13Dg->ChangeConcentrationBy(cell->GetPosition(), amount);
    }
  }
};

/////////////////////////////////////////////////////////////
//
// Define proliferation/growth behaviors:
//
/////////////////////////////////////////////////////////////

struct F_Proliferate : public Behavior {
  BDM_BEHAVIOR_HEADER(F_Proliferate, Behavior, 1);

  F_Proliferate() { AlwaysCopyToNew(); }

  virtual ~F_Proliferate() {}


  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 4;
      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();

      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      double TGFb_acLocalConc =
          tgfb_acDg->GetConcentration(cell->GetPosition());
      static auto* il13Dg = rm->GetDiffusionGrid(kIL13);
      double IL13LocalConc = il13Dg->GetConcentration(cell->GetPosition());
      static auto* fgf2Dg = rm->GetDiffusionGrid(kFGF2);
      double FGF2LocalConc = fgf2Dg->GetConcentration(cell->GetPosition());
      double F_proliferation_IL13_TGFb = sparam->F_proliferation_IL13_TGFb;

      double TGFb_threshold =
          TGFb_acLocalConc / (TGFb_acLocalConc + sparam->k_TGFb_ac);
      double IL13_threshold = IL13LocalConc / (IL13LocalConc + sparam->k_IL13);
      double FGF2_threshold = FGF2LocalConc / (FGF2LocalConc + sparam->k_FGF2);

      // Constant production + variable production (depends on TGFb, IL13 and
      // FGF2)
      double F_proliferate_threshold = F_proliferation_IL13_TGFb *
                                       (TGFb_threshold + IL13_threshold) *
                                       FGF2_threshold;

      F_proliferate_threshold /= (double)seconds_per_day_;

      cell->updateCurrentCellNumber(cell_type, 1 + F_proliferate_threshold);
    }
  }
};

struct AEC2_Proliferate : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_Proliferate, Behavior, 1);

  AEC2_Proliferate() { AlwaysCopyToNew(); }

  virtual ~AEC2_Proliferate() {}

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      size_t cell_type = 2;
      auto* sparam = Simulation::GetActive()->GetParam()->Get<SimParam>();

      double AEC2_proliferationRate =
          sparam->AEC2_proliferation / (double)seconds_per_day_;  // s^(-1)
      cell->updateCurrentCellNumber(cell_type, 1 + AEC2_proliferationRate);
    }
  }
};

/////////////////////////////////////////////////////////////
//
// Define apoptosis behaviors:
//
/////////////////////////////////////////////////////////////

struct Apoptosis : public Behavior {
  BDM_BEHAVIOR_HEADER(Apoptosis, Behavior, 1);

  Apoptosis() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();

      // First we update the AEC2 death
      double delta = sparam->ROS_death_AEC2/(double)seconds_per_day_;  // Increased apoptosis caused by
                                              // oxidative stress
      auto* rm = sim->GetResourceManager();
      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      double TGFb_acLocalConc =
          tgfb_acDg->GetConcentration(cell->GetPosition());
      double TGFb_threshold =
          (TGFb_acLocalConc / (TGFb_acLocalConc + sparam->k_TGFb_ac));
      double TGFb_ac_DeathRate =
          sparam->TGFb_ac_death_AEC2 / (double)seconds_per_day_;  // s^(-1)
          double additionalDeathRate =
              (1 + delta +
               TGFb_ac_DeathRate *
                   TGFb_threshold);

      // Now we remove cells based on their specific death rate
      cell->updateCurrentCellNumber(1, 1 - cell->getCellDeathRate(1));
      cell->updateCurrentCellNumber(2, 1 - (cell->getCellDeathRate(2)*additionalDeathRate));
      cell->updateCurrentCellNumber(4, 1 - cell->getCellDeathRate(4));
      cell->updateCurrentCellNumber(5, 1 - cell->getCellDeathRate(5));
      cell->updateCurrentCellNumber(7, 1 - cell->getCellDeathRate(7));
      cell->updateCurrentCellNumber(11, 1 - cell->getCellDeathRate(11));
      cell->updateCurrentCellNumber(14, 1 - cell->getCellDeathRate(14));
    }
  }
};

/////////////////////////////////////////////////////////////
//
// Define damage/repair behaviors:
//
/////////////////////////////////////////////////////////////

struct AEC2_Activation : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_Activation, Behavior, 1);

  AEC2_Activation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      int old_type_ = 13, new_type_ = 14;  // Damaged to activated

      double activationProbability = sparam->activationProbability_AEC2 /
                                     (double)seconds_per_day_;  // s^(-1)

      double current_AEC2_damaged = cell->getCurrentCellNumber(old_type_);
      double current_AEC2_activated = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(
          old_type_, current_AEC2_damaged * (1 - activationProbability));
      cell->setCurrentCellNumber(
          new_type_, current_AEC2_activated +
                         current_AEC2_damaged * activationProbability);
    }
  }
};

struct AEC2_Repair : public Behavior {  // Damaged cells (i.e. in damage zone)
                                        // not yet activated
  BDM_BEHAVIOR_HEADER(AEC2_Repair, Behavior, 1);

  AEC2_Repair() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      int old_type_ = 13, new_type_ = 2;  // Damaged to healthy
      double fibrocytesRepairRate =
          sparam->fibrocytesRepair_AEC2 / (double)seconds_per_day_;  // s^(-1)

      double current_AEC2_damaged = cell->getCurrentCellNumber(old_type_);
      double current_AEC2 = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(
          old_type_, current_AEC2_damaged * (1 - fibrocytesRepairRate));
      cell->setCurrentCellNumber(
          new_type_,
          current_AEC2 + current_AEC2_damaged * fibrocytesRepairRate);
    }
  }
};

/////////////////////////////////////////////////////////////
//
// Define differentiation behaviors:
//
/////////////////////////////////////////////////////////////

struct F_MF_Differentiation : public Behavior {
  BDM_BEHAVIOR_HEADER(F_MF_Differentiation, Behavior, 1);

  F_MF_Differentiation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      int old_type_ = 4, new_type_ = 11;

      auto* sim = Simulation::GetActive();
      auto* rm = sim->GetResourceManager();
      auto* sparam = sim->GetParam()->Get<SimParam>();

      static auto* tgfb_acDg = rm->GetDiffusionGrid(kTGFb_ac);
      static auto* pdgfDg = rm->GetDiffusionGrid(kPDGF);
      double tgfb_acLocalConc =
          tgfb_acDg->GetConcentration(cell->GetPosition());
      double pdgfLocalConc = pdgfDg->GetConcentration(cell->GetPosition());

      double transitionTGFb_ac = sparam->F_MF_TGFb_ac_differentiation *
                                 tgfb_acLocalConc /
                                 (tgfb_acLocalConc + sparam->k_TGFb_ac);

      transitionTGFb_ac /= (double)seconds_per_day_;  // s^(-1)

      double transitionPDGF = sparam->F_MF_PDGF_differentiation *
                              pdgfLocalConc / (pdgfLocalConc + sparam->k_PDGF);

      transitionPDGF /= (double)seconds_per_day_;  // s^(-1)

      double F_MF_transition_prob = transitionTGFb_ac + transitionPDGF;

      double current_F = cell->getCurrentCellNumber(old_type_);
      double current_MF = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(old_type_,
                                 current_F * (1 - F_MF_transition_prob));
      cell->setCurrentCellNumber(new_type_,
                                 current_MF + current_F * F_MF_transition_prob);
    }
  }
};

struct MO1_Differentiation : public Behavior {
  BDM_BEHAVIOR_HEADER(MO1_Differentiation, Behavior, 1);

  MO1_Differentiation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      int old_type_ = 12, new_type_ = 7;
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      auto* rm = sim->GetResourceManager();

      // Monocyte flux part
      auto* mcp1Dg = rm->GetDiffusionGrid(kMCP1);

      double MCP1conc = mcp1Dg->GetConcentration(cell->GetPosition()),
             basic_prob = 0.0;

      double current_M0 = cell->getCurrentCellNumber(old_type_);
      double current_M1 = cell->getCurrentCellNumber(new_type_);

      double MCP1_threshold = MCP1conc / (MCP1conc + sparam->k_MCP1);

      // If MCP-1 threshold is lower than the basic production probability for
      // M0, we use the latter. In this way we ensure that even in homeostasis
      // we keep a steady level of M1
      if (MCP1_threshold < (sparam->M0_basic_prob_production)) {
        basic_prob = sparam->M0_basic_prob_production;
      } else {
        basic_prob = MCP1_threshold;
      }

// beta and basic_prob can act as a single term
      double M1_from_M0 =
          (basic_prob * (sparam->beta) * (current_M0 - current_M1)) /
          (double)seconds_per_day_;

      // We don't update the number of M0, is fixed as it's in the paper
      cell->setCurrentCellNumber(new_type_, current_M1 + M1_from_M0);
    }
  }
};

struct M12_Differentiation : public Behavior {
  BDM_BEHAVIOR_HEADER(M12_Differentiation, Behavior, 1);

  M12_Differentiation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();

      int old_type_ = 7, new_type_ = 5;
      double M1differentiationProb =
          sparam->M1_M2_differentiation / (double)seconds_per_day_;  // s^(-1)

      double current_M1 = cell->getCurrentCellNumber(old_type_);
      double current_M2 = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(old_type_,
                                 current_M1 * (1 - M1differentiationProb));
      cell->setCurrentCellNumber(
          new_type_, current_M2 + current_M1 * M1differentiationProb);
    }
  }
};

struct M21_Differentiation : public Behavior {
  BDM_BEHAVIOR_HEADER(M21_Differentiation, Behavior, 1);

  M21_Differentiation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      auto* rm = sim->GetResourceManager();

      int old_type_ = 5, new_type_ = 7;
      static auto* tnfaDg = rm->GetDiffusionGrid(kTNFa);
      double tnfaLocalConc = tnfaDg->GetConcentration(cell->GetPosition());
      double transitionTNFa =
          sparam->M2_M1_TNFa_differentiation *
          (tnfaLocalConc / (tnfaLocalConc + sparam->k_TNFa));
      transitionTNFa /= (double)seconds_per_day_;  // s^(-1)

      double current_M2 = cell->getCurrentCellNumber(old_type_);
      double current_M1 = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(old_type_, current_M2 * (1 - transitionTNFa));
      cell->setCurrentCellNumber(new_type_,
                                 current_M1 + current_M2 * transitionTNFa);

      // TNFa depletion by M2 (0.55*10^-3 is the nb of M2 in the Hao paper)
      double TNFa_depletion =
          transitionTNFa * 0.00055 *
          (current_M2 / (cell->getInitialCellNumber(old_type_)));
      if (TNFa_depletion < tnfaLocalConc) {
        tnfaDg->ChangeConcentrationBy(cell->GetPosition(), -TNFa_depletion);
      } else {
        tnfaDg->ChangeConcentrationBy(cell->GetPosition(), -tnfaLocalConc);
      }
    }
  }
};

struct AEC21_Differentiation : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC21_Differentiation, Behavior, 1);

  AEC21_Differentiation() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      int old_type_ = 2, new_type_ = 1;
      auto* sparam = sim->GetParam()->Get<SimParam>();
      double AEC1_productionRate =
          sparam->AEC2_AEC1_differentiation / (double)seconds_per_day_;

      double current_AEC2 = cell->getCurrentCellNumber(old_type_);
      double current_AEC1 = cell->getCurrentCellNumber(new_type_);

      cell->setCurrentCellNumber(old_type_,
                                 current_AEC2 * (1 - AEC1_productionRate));
      cell->setCurrentCellNumber(
          new_type_, current_AEC1 + current_AEC2 * AEC1_productionRate);
    }
  }
};

struct AEC2_EMT : public Behavior {
  BDM_BEHAVIOR_HEADER(AEC2_EMT, Behavior, 1);

  AEC2_EMT() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      int old_type_ = 14, new_type_1_ = 4, new_type_2_ = 11;

      double AEC2_EMT_rate =
          sparam->AEC2_EMT_constant_rate / (double)seconds_per_day_;

      double AEC2_EMT_F_rate =
          sparam->AEC2_F;  // Transition to fibroblast, no need for s^-1
      double AEC2_EMT_MF_rate =
          1 - AEC2_EMT_F_rate;  // Damaged AEC2 which go through EMT and don't
                                // differentiate into F, differentiate into MF

      double current_AEC2_activated = cell->getCurrentCellNumber(old_type_);
      double current_F = cell->getCurrentCellNumber(new_type_1_);
      double current_MF = cell->getCurrentCellNumber(new_type_2_);

      cell->setCurrentCellNumber(old_type_,
                                 current_AEC2_activated * (1 - AEC2_EMT_rate));
      cell->setCurrentCellNumber(
          new_type_1_,
          current_F + current_AEC2_activated * AEC2_EMT_rate * AEC2_EMT_F_rate);
      cell->setCurrentCellNumber(
          new_type_2_, current_MF + current_AEC2_activated * AEC2_EMT_rate *
                                        AEC2_EMT_MF_rate);
    }
  }
};

/////////////////////////////////////////////////////////////
//
// Define add cells behaviors (old operations):
//
/////////////////////////////////////////////////////////////

struct F_Addition : public Behavior {
  BDM_BEHAVIOR_HEADER(F_Addition, Behavior, 1);

  F_Addition() { AlwaysCopyToNew(); }

  void Run(Agent* so) override {
    if (auto* cell = bdm_static_cast<alveolarDuct*>(so)) {
      auto* sim = Simulation::GetActive();
      auto* sparam = sim->GetParam()->Get<SimParam>();
      int old_type_ = 2, new_type_ = 4;

      double current_AEC2 = cell->getCurrentCellNumber(old_type_);
      double current_F = cell->getCurrentCellNumber(new_type_);

      double new_F_probability =
          sparam->F_production / (double)seconds_per_day_;

      // AEC2 are not removed, their number is only used to compute the number
      // of F to be added as this depends on the AEC2 conc
      cell->setCurrentCellNumber(new_type_,
                                 current_F + current_AEC2 * new_F_probability);
    }
  }
};


}  // namespace bdm

#endif  // BIOLOGY_MODULES_H_
