// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#include "geometry.h"

namespace bdm {

// Linker needs this re-declarations
std::vector<std::vector<Double3>> endoPositionsAlv;
std::vector<int> segmentIDPerAlveolus;
std::vector<Double3> alveolarProjsOnAxis;

double const distanceBetweenPoints(Double3 const point1, Double3 const point2) {
  double dx = point1[0] - point2[0];
  double dy = point1[1] - point2[1];
  double dz = point1[2] - point2[2];
  double distance = sqrt(dx * dx + dy * dy + dz * dz);

  return distance;
}

// If the end of the segment (point) which is perpendicular to a segment (line)
// doesn't lie on it, return the end of the segment which is closer to that
// point (i. e. clamp the point to the segment)
Double3 ClampPointToLine(Double3 point, segment line) {
  double minX, minY, minZ, maxX, maxY, maxZ;
  Double3 clampedPoint;  // new coordinates of the clamped point

  // set max and min x, y, z based on their values (makes it easier to compare
  // later)
  if (line.origin_[0] <= line.end_[0]) {
    minX = line.origin_[0];
    maxX = line.end_[0];
  } else {
    minX = line.end_[0];
    maxX = line.origin_[0];
  }
  if (line.origin_[1] <= line.end_[1]) {
    minY = line.origin_[1];
    maxY = line.end_[1];
  } else {
    minY = line.end_[1];
    maxY = line.origin_[1];
  }
  if (line.origin_[2] <= line.end_[2]) {
    minZ = line.origin_[2];
    maxZ = line.end_[2];
  } else {
    minZ = line.end_[2];
    maxZ = line.origin_[2];
  }

  // if the point lies on the segment just return it, otherwise return the end
  // of segment which is closer to it
  clampedPoint[0] = (point[0] < minX)   ? minX
                    : (point[0] > maxX) ? maxX
                                        : point[0];
  clampedPoint[1] = (point[1] < minY)   ? minY
                    : (point[1] > maxY) ? maxY
                                        : point[1];
  clampedPoint[2] = (point[2] < minZ)   ? minZ
                    : (point[2] > maxZ) ? maxZ
                                        : point[2];

  return clampedPoint;
}

// given two segment (cylinders), check if their distance is lower than the sum
// of their radii. The distance between them is the length of the segment that
// is perpendicular to both the segments
bool segmentsColliding(segment first, segment second, double radius) {
  // the first segment starts at p1
  double p1x = first.origin_[0], p1y = first.origin_[1], p1z = first.origin_[2];
  // the second segment starts at p3
  double p3x = second.origin_[0], p3y = second.origin_[1],
         p3z = second.origin_[2];
  // d1 is the direction of the first segment (not normalized)
  double d1x = first.segmentDirection[0], d1y = first.segmentDirection[1],
         d1z = first.segmentDirection[2];
  // d2 is the direction of the second segment (not normalized)
  double d2x = second.segmentDirection[0], d2y = second.segmentDirection[1],
         d2z = second.segmentDirection[2];

  // A...F are coefficients that come from calculations, imposing that i1 (the
  // first end of the perpendicular segment) lies on the first segment and i2
  // (the second end of the perpendicular segment) lies on the second segment
  double A =
      d1x * p3x - d1x * p1x + d1y * p3y - d1y * p1y + d1z * p3z - d1z * p1z;
  double B = d1x * d2x + d1y * d2y + d1z * d2z;
  double C = -d1x * d1x - d1y * d1y - d1z * d1z;
  double D =
      d2x * p3x - d2x * p1x + d2y * p3y - d2y * p1y + d2z * p3z - d2z * p1z;
  double E = d2x * d2x + d2y * d2y + d2z * d2z;
  double F = -d2x * d1x - d2y * d1y - d2z * d1z;

  // m and n are the coefficients used to determine the coordinates of i1 and i2
  double m = ((A * E / B) - D) / (F - (C * E / B));
  double n = (-m * C - A) / B;

  // we impose that i1 lies on line_1 and i2 lies on line_2
  Double3 i1 = first.origin_ + first.segmentDirection * m;
  Double3 i2 = second.origin_ + second.segmentDirection * n;

  // we constrain the range of the two ends of the perpendicular segment to be
  // on the two segments
  Double3 i1Clamped = ClampPointToLine(i1, first);
  Double3 i2Clamped = ClampPointToLine(i2, second);

  // distance is the length of the perpendicular segment
  double diClamped_x = i1Clamped[0] - i2Clamped[0];
  double diClamped_y = i1Clamped[1] - i2Clamped[1];
  double diClamped_z = i1Clamped[2] - i2Clamped[2];
  double distance = sqrt(diClamped_x * diClamped_x + diClamped_y * diClamped_y +
                         diClamped_z * diClamped_z);

  // since radius is the radius of the segment (from the axis to the
  // circumference where the centers of the alveoli lie), we must also take into
  // account the radius of the alveolus located at its end (and sum it)
  return distance < ((radius + alveolar_radius) * 2);
}

// Compute distance vector delta between two points B and A (in cartesian
// coordinates)
Double3 deltaCartesianCoords(coord B, coord A) {
  Double3 delta = {{B.x - A.x, B.y - A.y, B.z - A.z}};
  return delta;
}

// rotate vector by thetaRot around a given axis passing through center
Double3 rotateVector(Double3 oldCoordinates, Double3 axis, double thetaRot,
                     Double3 center) {
  double u = axis[0], v = axis[1], w = axis[2];
  thetaRot = thetaRot * PI / straight;
  double u2 = u*u, v2 = v*v, w2 = w*w;
  double cosT = cos(thetaRot), sinT = sin(thetaRot);
  double a = center[0], b = center[1], c = center[2];

  Double4 rotationMatrix[4];

  rotationMatrix[0][0] = u2 + (v2 + w2) * cosT;
  rotationMatrix[0][1] = u * v * (1 - cosT) - w * sinT;
  rotationMatrix[0][2] = u * w * (1 - cosT) + v * sinT;
  rotationMatrix[0][3] = (a * (v2 + w2) - u * (b * v + c * w)) * (1 - cosT) +
                         (b * w - c * v) * sinT;
  rotationMatrix[1][0] = u * v * (1 - cosT) + w * sinT;
  rotationMatrix[1][1] = v2 + (u2 + w2) * cosT;
  rotationMatrix[1][2] = v * w * (1 - cosT) - u * sinT;
  rotationMatrix[1][3] = (b * (u2 + w2) - v * (a * u + c * w)) * (1 - cosT) +
                         (c * u - a * w) * sinT;
  rotationMatrix[2][0] = u * w * (1 - cosT) - v * sinT;
  rotationMatrix[2][1] = v * w * (1 - cosT) + u * sinT;
  rotationMatrix[2][2] = w2 + (u2 + v2) * cosT;
  rotationMatrix[2][3] = (c * (u2 + v2) - w * (a * u + b * v)) * (1 - cosT) +
                         (a * v - b * u) * sinT;
  rotationMatrix[3][0] = 0;
  rotationMatrix[3][1] = 0;
  rotationMatrix[3][2] = 0;
  rotationMatrix[3][3] = 1;

  TMatrixD mat(4, 4);
  TArrayD data(16);

  // Move data from Double4 [4] to TArrayD
  int i = 0;

  for (Int_t col = 0; col < 4; col++) {
    for (Int_t row = 0; row < 4; row++) {
      data[i] = rotationMatrix[col][row];
      i++;
    }
  }

  // Generate TMatrixD from TArrayD
  mat.SetMatrixArray(data.GetArray());

  TArrayD vec_data(oldCoordinates.size() + 1);

  for (size_t i = 0; i < oldCoordinates.size(); i++) {
    vec_data[i] = oldCoordinates[i];
  }

  vec_data[3] = 1.;  // Last element must be 1

  TMatrixD oldVec(4, 1), newVec(4, 1);

  // Get 4x1 matrix from TArrayD
  oldVec.SetMatrixArray(vec_data.GetArray());

  // Compute new coordinates
  newVec = mat * oldVec;

  double const* start = newVec.GetMatrixArray();

  Double3 newCoordinates = {start[0], start[1], start[2]};

  return newCoordinates;
}

//(for regular structure only) return gamma angle (wrt father axis) given the
// index
double getGamma(int index) {
  switch (index) {  // for the first 14 indexes (until generation 4) set gamma
                    // to 45 degress
    case 1 ... 30:
      return 45.;  // 60.

    default:  // then set it randomly (do not exceed 0-89 range since the
              // segment axis rotation is not implemented for values outside
              // that range)
      auto* random = Simulation::GetActive()->GetRandom();
      return random->Uniform(45., 89.);
  }
}

//(for regular structure only) return psi angle (rotation of XY plane of the son
// with respect to XY plane of the father) given the
// index
double getPsi(int index) {
  auto* random = Simulation::GetActive()->GetRandom();

  // set segment 1 and 2 at 180 degress from each other
  if (index == 1) {
    return 0.0;
  } else {
    if (index == 2) {
      return 180.0;
    } else if (index % 2 != 0 &&
               index <= 30) {  // set segments between 3 and 14 at 90 degrees
                               // from each other (45 degrees for branch 1, 45 +
                               // 270 = 315 for branch 2)
      return 90.0;
    } else if (index % 2 == 0 && index <= 30) {
      return 270.0;
    } else if (index % 2 != 0 &&
               index >
                   30) {  // segments belonging to generations after the fourth
                          // get random psi (branch 1 and 2 must not overlap)
      return random->Uniform(20., 160.0);
    } else {
      return random->Uniform(200.0, 340.0);
    }
  }
}

TMatrixD getRotationMatrix(Double3 center, Double3 z_from, Double3 z_to,
                           Double3 x_from, Double3 x_to, Double3 y_from,
                           Double3 y_to) {
  // Generate local axes
  Double3 x_axis = x_to - x_from;
  Double3 y_axis = y_to - y_from;
  Double3 z_axis = z_to - z_from;

  x_axis.Normalize();
  y_axis.Normalize();
  z_axis.Normalize();

  Double4 rotation_matrix[4];

  // Generate rotation matrix in the form
  /*
  | x_towardsX  x_towardsY  x_towardsZ  center_x  |
  | y_towardsX  y_towardsY  y_towardsZ  center_y  |
  | z_towardsX  z_towardsY  z_towardsZ  center_z  |
  |      0           0           0         1      |
  */
  rotation_matrix[0] = {x_axis[0], x_axis[1], x_axis[2], 0};
  rotation_matrix[1] = {y_axis[0], y_axis[1], y_axis[2], 0};
  rotation_matrix[2] = {z_axis[0], z_axis[1], z_axis[2], 0};
  rotation_matrix[3] = {center[0], center[1], center[2], 1};

  TMatrixD mat(4, 4);
  TArrayD data(16);

  // Move data from Double4 [4] to TArrayD
  int i = 0;

  for (Int_t col = 0; col < 4; col++) {
    for (Int_t row = 0; row < 4; row++) {
      data[i] = rotation_matrix[col][row];
      i++;
    }
  }

  // Generate TMatrixD from TArrayD
  mat.SetMatrixArray(data.GetArray(), "F");

  // To be used when converting from gloval coords to local coords
  // TMatrixD invMat(4, 4)
  // invMat = mat.Invert();

  return mat;
}

// Pass as arguments local coords and change matrix to get the global coords,
// global coords and inverse of the change matrix to get the local coords
Double3 changeCoordinates(Double3 oldCoordinates, TMatrixD changeMatrix) {
  TArrayD data(oldCoordinates.size() + 1);

  for (size_t i = 0; i < oldCoordinates.size(); i++) {
    data[i] = oldCoordinates[i];
  }

  data[3] = 1.;  // Last element must be 1

  TMatrixD oldVec(4, 1), newVec(4, 1);

  // Get 4x1 matrix from TArrayD
  oldVec.SetMatrixArray(data.GetArray());

  // Compute new coordinates
  newVec = changeMatrix * oldVec;

  const double* start;

  start = newVec.GetMatrixArray();

  Double3 newCoordinates;
  newCoordinates = {{start[0], start[1], start[2]}};

  // Return new coordinates
  return newCoordinates;
}

// Compute cross product between two 3d vectors
Double3 crossProduct(Double3 vec1, Double3 vec2) {
  Double3 vec3;
  vec3[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
  vec3[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
  vec3[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

  return vec3;
}

Double3 rotateXYonPlane(Double3 oldCoords, double theta) {
  Double3 newCoords;

  newCoords[0] = oldCoords[0] * cos(theta * PI / straight) -
                 oldCoords[1] * sin(theta * PI / straight);
  newCoords[1] = oldCoords[0] * sin(theta * PI / straight) +
                 oldCoords[1] * cos(theta * PI / straight);
  newCoords[2] = oldCoords[2];

  return newCoords;
}

// Compute distance between a point and a segment (from Wolfram Alpha)
double pointToSegmentDistance(Double3 point, Double3 segmentOrigin,
                              Double3 segmentEnd) {
  double d = (crossProduct(point - segmentOrigin, point - segmentEnd).Norm()) /
             ((segmentEnd - segmentOrigin).Norm());

  return d;
}

  void segmentStructure(segment* Segments) {
    Log::Warning("", mean_segment_length);
    int const numberOfSegments =
        pow(2, generations) - 1;  // sum of 2^i for i from 0 to 9 (10 generations)

    for (int i = 1; i < numberOfSegments; i += 2) {
      // index of the father is index of the son (branch) +1 divided by 2
      // (branches 2 and 3 are sons of 1). -1 since array indexes start from 0
      int father = ((i + 1) / 2) - 1;

      // Projects father's endpoints along its axis by meansegmentlength. Those
      // will be used to determine the new branches' endpoints
      Double3 fathersEndProjected =
          Segments[father].end_ +
          Segments[father].segmentNormDirection * mean_segment_length;
      Double3 fathers_X_in_Projected =
          Segments[father].x_in +
          Segments[father].segmentNormDirection * mean_segment_length;
      Double3 fathers_X_fin_Projected =
          Segments[father].x_fin +
          Segments[father].segmentNormDirection * mean_segment_length;
      Double3 fathers_Y_in_Projected =
          Segments[father].y_in +
          Segments[father].segmentNormDirection * mean_segment_length;
      Double3 fathers_Y_fin_Projected =
          Segments[father].y_fin +
          Segments[father].segmentNormDirection * mean_segment_length;
      Double3 fathers_center_Projected =
          Segments[father].center +
          Segments[father].segmentNormDirection * mean_segment_length;

      // Matrix used to get global coordinates for the new branches. Translated
      // father's endpoints are used to generate it
      TMatrixD rotMatrix(4, 4);
      rotMatrix = getRotationMatrix(
          fathers_center_Projected, Segments[father].end_, fathersEndProjected,
          fathers_X_in_Projected, fathers_X_fin_Projected, fathers_Y_in_Projected,
          fathers_Y_fin_Projected);

      // Branch 1
      Segments[i].segmentID_ =
          i + 1;  // the first segment has index 1 and so on...

      bool intersection = true;

      int counter1 = 0;

      while (intersection) {  // loop until there's no intersection between
                              // segment i and the others

        Segments[i] = segment();  // Clean the structure

        // Cylinder's axis is used to rotate the local coordinates system (of the
        // translated father) by psi
        Double3 rotationAxis_b1 =
            (Segments[i].end_ - Segments[i].origin_).Normalize();

        Segments[i].psi_ = getPsi(i);

        Segments[i].x_in = rotateVector(Segments[i].x_in, rotationAxis_b1,
                                        Segments[i].psi_, Segments[i].center);
        Segments[i].x_fin = rotateVector(Segments[i].x_fin, rotationAxis_b1,
                                         Segments[i].psi_, Segments[i].center);
        Segments[i].y_in = rotateVector(Segments[i].y_in, rotationAxis_b1,
                                        Segments[i].psi_, Segments[i].center);
        Segments[i].y_fin = rotateVector(Segments[i].y_fin, rotationAxis_b1,
                                         Segments[i].psi_, Segments[i].center);

        // Change rotated and translated coordinates from local to global
        Segments[i].x_in = changeCoordinates(Segments[i].x_in, rotMatrix);
        Segments[i].x_fin = changeCoordinates(Segments[i].x_fin, rotMatrix);
        Segments[i].y_in = changeCoordinates(Segments[i].y_in, rotMatrix);
        Segments[i].y_fin = changeCoordinates(Segments[i].y_fin, rotMatrix);
        Segments[i].center = changeCoordinates(Segments[i].center, rotMatrix);
        Segments[i].origin_ = changeCoordinates(Segments[i].origin_, rotMatrix);
        Segments[i].end_ = changeCoordinates(Segments[i].end_, rotMatrix);

        Segments[i].gamma_ = getGamma(i);

        // Use branch y axis at the top of the branch (i.e. crossing its origin)
        // to rotate the branch by gamma with respect to its father's axis The
        // direction of the rotation is given by the positive direction of the x
        // axis of the branch
        Double3 rotAxIn =
            Segments[i].y_in -
            Segments[father].segmentNormDirection * mean_segment_length / 2.;
        Double3 rotAxFin =
            Segments[i].y_fin -
            Segments[father].segmentNormDirection * mean_segment_length / 2.;

        rotationAxis_b1 = (rotAxFin - rotAxIn).Normalize();

        Segments[i].x_in = rotateVector(Segments[i].x_in, rotationAxis_b1,
                                        -Segments[i].gamma_, Segments[i].origin_);
        Segments[i].x_fin =
            rotateVector(Segments[i].x_fin, rotationAxis_b1, -Segments[i].gamma_,
                         Segments[i].origin_);
        Segments[i].y_in = rotateVector(Segments[i].y_in, rotationAxis_b1,
                                        -Segments[i].gamma_, Segments[i].origin_);
        Segments[i].y_fin =
            rotateVector(Segments[i].y_fin, rotationAxis_b1, -Segments[i].gamma_,
                         Segments[i].origin_);
        Segments[i].end_ = rotateVector(Segments[i].end_, rotationAxis_b1,
                                        -Segments[i].gamma_, Segments[i].origin_);
        Segments[i].center =
            rotateVector(Segments[i].center, rotationAxis_b1, -Segments[i].gamma_,
                         Segments[i].origin_);

        // Compute useful quantities
        Segments[i].segmentDirection = (Segments[i].end_ - Segments[i].origin_);
        Segments[i].segmentNormDirection =
            Segments[i].segmentDirection.Normalize();

        // Segment collision
        bool check = 1;  // if check doesn't change, segments doesn't collide with
                         // other segments in the structure

        if (i > 14) {  // don't check the first 4 generations since their
                       // arrangement is set by the user (regular)
          for (int s = 15; s < i; s++) {  // check all the segments but the ones
                                          // that belong to generations 1-4
            if (s != father &&
                segmentsColliding(
                    Segments[i],
                    Segments[s],  // do not check collision with father
                    segmentRadius)) {
              check = 0;
              break;  // exit as soon as an intersection is found
            }
          }
        }

        if (check == 1) {
          intersection = 0;  // check = 1 means no intersection
        }

        counter1++;
        if (counter1 > 10000) {
          i = 1;
          break;
        }
      }

      if (counter1 > 10000) {
        continue;
      }

      // Branch 2 (same as branch 1 but using i+1 and do not check collision with
      // branch 1)

      Segments[i + 1].segmentID_ =
          i + 2;  // the first segment has index 1 and so on...

      intersection = true;

      int counter2 = 0;

      while (intersection) {
        Segments[i + 1] = segment();  // Clean the structure

        Double3 rotationAxis_b2 =
            (Segments[i + 1].end_ - Segments[i + 1].origin_).Normalize();

        Segments[i + 1].psi_ = getPsi(i + 1);

        Segments[i + 1].x_in =
            rotateVector(Segments[i + 1].x_in, rotationAxis_b2,
                         Segments[i + 1].psi_, Segments[i + 1].center);
        Segments[i + 1].x_fin =
            rotateVector(Segments[i + 1].x_fin, rotationAxis_b2,
                         Segments[i + 1].psi_, Segments[i + 1].center);
        Segments[i + 1].y_in =
            rotateVector(Segments[i + 1].y_in, rotationAxis_b2,
                         Segments[i + 1].psi_, Segments[i + 1].center);
        Segments[i + 1].y_fin =
            rotateVector(Segments[i + 1].y_fin, rotationAxis_b2,
                         Segments[i + 1].psi_, Segments[i + 1].center);

        Segments[i + 1].x_in = changeCoordinates(Segments[i + 1].x_in, rotMatrix);
        Segments[i + 1].x_fin =
            changeCoordinates(Segments[i + 1].x_fin, rotMatrix);
        Segments[i + 1].y_in = changeCoordinates(Segments[i + 1].y_in, rotMatrix);
        Segments[i + 1].y_fin =
            changeCoordinates(Segments[i + 1].y_fin, rotMatrix);
        Segments[i + 1].center =
            changeCoordinates(Segments[i + 1].center, rotMatrix);
        Segments[i + 1].origin_ =
            changeCoordinates(Segments[i + 1].origin_, rotMatrix);
        Segments[i + 1].end_ = changeCoordinates(Segments[i + 1].end_, rotMatrix);

        Segments[i + 1].gamma_ = getGamma(i + 1);

        Double3 rotAxIn =
            Segments[i + 1].y_in -
            Segments[father].segmentNormDirection * mean_segment_length / 2.;
        Double3 rotAxFin =
            Segments[i + 1].y_fin -
            Segments[father].segmentNormDirection * mean_segment_length / 2.;

        rotationAxis_b2 = (rotAxFin - rotAxIn).Normalize();

        Segments[i + 1].x_in =
            rotateVector(Segments[i + 1].x_in, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);
        Segments[i + 1].x_fin =
            rotateVector(Segments[i + 1].x_fin, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);
        Segments[i + 1].y_in =
            rotateVector(Segments[i + 1].y_in, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);
        Segments[i + 1].y_fin =
            rotateVector(Segments[i + 1].y_fin, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);
        Segments[i + 1].end_ =
            rotateVector(Segments[i + 1].end_, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);
        Segments[i + 1].center =
            rotateVector(Segments[i + 1].center, rotationAxis_b2,
                         -Segments[i + 1].gamma_, Segments[i + 1].origin_);

        Segments[i + 1].segmentDirection =
            (Segments[i + 1].end_ - Segments[i + 1].origin_);
        Segments[i + 1].segmentNormDirection =
            Segments[i + 1].segmentDirection.Normalize();

        bool check = 1;

        if (i + 1 > 14) {
          for (int s = 15; s < i - 1; s++) {
            if (s != father &&
                segmentsColliding(Segments[i + 1], Segments[s], segmentRadius)) {
              check = 0;
              break;
            }
          }
        }

        if (check == 1) {
          intersection = 0;
        }

        counter2++;
        if (counter2 > 10000) {
          i = 1;
          break;
        }
      }


    }  // Segments loop
  }

}  // namespace bdm
