// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#ifndef GEOMETRY_H_
#define GEOMETRY_H_

#include <math.h>
#include <vector>
#include "TArrayD.h"
#include "TMatrixD.h"
#include "biodynamo.h"

namespace bdm {

// Generic constants
double const PI = 3.14159265;
double const turn = 360.0;
double const straight = 180.0;
double const waterDensity = 1.03;       // g/ml
double const um3_To_ml = pow(10, -12);  // 1 um^3 = 10^-12 ml
double const um3_To_cm3 = um3_To_ml;

// Geometric constants
double const alveolar_radius = 110.16;  // um
double const alveolar_reduced_radius =
    alveolar_radius * cos(60 * PI / straight);
double const alveolar_extended_radius =
    alveolar_radius + 8.295;  // Value hardcoded is AECI radius, needed to
                              // overcome superposition of alveoli
double const alveolar_extended_diameter = alveolar_extended_radius * 2;
double const alveolar_surface =
    3.0 * PI * alveolar_radius*alveolar_radius;        // um^2, 3/4 of a sphere
double const interstitial_space_width = 0.54;  // um
int const nb_on_xy_plane = 6; /*1*/ /*6*/      // Actual number used
int const nb_longitudinal = 3; /*1*/           /*3, 4 to fill the junctions*/
int const max_nb_on_xy_plane =
    6;  // Used for segmentRadius calculation, with this results are closer to
        // those from Weibel
double const segmentRadius =
    alveolar_extended_diameter /
    (2 * sin(PI / static_cast<double>(max_nb_on_xy_plane)));

static double segment_internal_radius = segmentRadius - alveolar_reduced_radius;

size_t const alveoli_per_segment =
    nb_on_xy_plane * nb_longitudinal; /*18 (3z*8xy) or 3 for testing*/
int const generations = 7;  // In each acinus (dichotomous): generations from
                            // 0 to 9 (10 in total, works until 6)
size_t const nb_of_segments =
    pow(2, generations) - 1;  // sum of 2^i for i from 0 to 9 (10 generations)
static size_t nb_of_alveoli = alveoli_per_segment * nb_of_segments;
int const numberOfProjections = nb_of_alveoli / nb_on_xy_plane;

double const mean_segment_length =
    nb_longitudinal * 2 * alveolar_extended_radius;  // 660um = 3 alveoli (3*d = 3*220um)

// First vector is alveolar index, second vector are endo cell positions for
// each alveolus. Position and number of endo cells are fixed, so we compute
// them here and then pass them to the monocyte flux op
extern std::vector<std::vector<Double3>> endoPositionsAlv;
extern std::vector<int> segmentIDPerAlveolus;
extern std::vector<Double3> alveolarProjsOnAxis;

// Provides cartesian and spherical coordinates as well as methods to convert
// between them
struct coord {
  double x = 0., y = 0., z = 0.;
  double theta = 0., phi = 0., ro = 0.;  // theta on XY plane, phi along z axis
  double diameter = 0.;

  // Setters and getters for coord
  void setCartesianCoords(double x_coord, double y_coord, double z_coord) {
    x = x_coord;
    y = y_coord;
    z = z_coord;
  }

  void setCartesianCoords(Double3 cartesianCoords) {
    x = cartesianCoords[0];
    y = cartesianCoords[1];
    z = cartesianCoords[2];
  }

  // theta in [0, 2PI], phi in [0, PI]
  void setSphericalCoords(Double3 sphericalCoords) {
    ro = sphericalCoords[0];
    theta = sphericalCoords[1];
    phi = sphericalCoords[2];
  }

  Double3 getSphericalCoords() { return {ro, theta, phi}; }

  Double3 getCartesianCoords() { return {x, y, z}; }

  void sphericalToCartesian(Double3 sphereCenter) {
    x = sphereCenter[0] +
        ro * sin(phi * PI / straight) * cos(theta * PI / straight);
    y = sphereCenter[1] +
        ro * sin(phi * PI / straight) * sin(theta * PI / straight);
    z = sphereCenter[2] + ro * cos(phi * PI / straight);
  }

  void cartesianToSpherical(Double3 sphereCenter) {
    double x_local = x - sphereCenter[0], y_local = y - sphereCenter[1],
           z_local = z - sphereCenter[2];
    ro = sqrt(x_local*x_local+y_local*y_local+z_local*z_local);
    theta = (atan2(y_local, x_local) * straight / PI);
    if (theta <= 0.) {
      theta += turn;
    }  // Convert from [-PI, PI] to [0, 2PI]
    phi = acos(z_local / ro) * straight / PI;
  }
  // Prevent theta from exiting the interval [0,360] (check -deltaTheta)
  void updateTheta(double deltaTheta) {
    double finalTheta = theta + deltaTheta;

    if (finalTheta >= 0. && finalTheta <= turn) {
      theta += deltaTheta;
    } else {
      if (finalTheta < 0.0) {  //+1 -2 = -1 -> 360 + (1+(-2))
        theta = turn + finalTheta;
      } else {
        theta = finalTheta -
                turn;  // theta+deltaTheta>360.0. 359+5 = 364 -> 364-360 = 4
      }
    }
  }

  // Prevent phi from exiting the interval [0,180]
  void updatePhi(double deltaPhi) {
    double finalPhi = phi + deltaPhi;

    if (finalPhi >= 0.0 && finalPhi <= straight) {
      phi += deltaPhi;
    } else {
      if (finalPhi < 0.0) {
        phi = straight + finalPhi;
      } else {
        phi = finalPhi - straight;  // phi+deltaPhi>180.0
      }
    }
  }
};

// structure for segments
struct segment {
  // Initialize as the first segment (which is vertical) so doesn't need to be
  // done later
  double psi_ = 0.0, gamma_ = 0.0;  // psi on XY plane, gamma along z axis
  Double3 origin_ = {0., 0., -mean_segment_length / 2.};  // starting point
  Double3 end_ = {0., 0., mean_segment_length / 2.};
  int segmentID_ = 1;
  Double3 segmentDirection = end_ - origin_;
  Double3 segmentNormDirection = segmentDirection.Normalize();
  Double3 alveoliCenters[alveoli_per_segment];
  int alveoliGlobalIndex[alveoli_per_segment];
  // Local coordinates
  Double3 x_in = {-segment_internal_radius, 0., 0.};
  Double3 x_fin = {segment_internal_radius, 0., 0.};
  Double3 y_in = {0., -segment_internal_radius, 0.};
  Double3 y_fin = {0., segment_internal_radius, 0.};
  Double3 center = {0., 0., 0.};
};

double const distanceBetweenPoints(Double3 const point1, Double3 const point2);

// If the end of the segment (point) which is perpendicular to a segment (line)
// doesn't lie on it, return the end of the segment which is closer to that
// point (i. e. clamp the point to the segment)
Double3 ClampPointToLine(Double3 point, segment line);

// given two segment (cylinders), check if their distance is lower than the sum
// of their radii. The distance between them is the length of the segment that
// is perpendicular to both the segments
bool segmentsColliding(segment first, segment second, double radius);

// Compute distance vector delta between two points B and A (in cartesian
// coordinates)
Double3 deltaCartesianCoords(coord B, coord A);

// rotate vector by thetaRot around a given axis
Double3 rotateVector(Double3 oldCoordinates, Double3 axis, double thetaRot,
                     Double3 center);

//(for regular structure only) return gamma angle (wrt z axis) given the index
double getGamma(int index);

//(for regular structure only) return psi angle (rotation on XY plane) given the
// index
double getPsi(int index);

// Generates segments starting points based on dichotomical structure.
void segmentStructure(segment* Segments);


TMatrixD getRotationMatrix(Double3 center, Double3 z_from, Double3 z_to,
                           Double3 x_from, Double3 x_to, Double3 y_from,
                           Double3 y_to);

Double3 changeCoordinates(Double3 oldCoordinates, TMatrixD changeMatrix);

Double3 crossProduct(Double3 vec1, Double3 vec2);

double pointToSegmentDistance(Double3 point, Double3 segmentOrigin,
                              Double3 segmentEnd);

}  // namespace bdm

#endif  // GEOMETRY_H_
