// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#ifndef MY_PARAM_H_
#define MY_PARAM_H_

#include "core/param/param_group.h"

namespace bdm {

/// This class defines parameters that are specific to this simulation.
struct SimParam : public ParamGroup {
  BDM_PARAM_GROUP_HEADER(SimParam, 1);

  // Operations
  size_t nb_of_sim_steps_ = 3600 * 2;
  size_t nb_of_subs_ = 10;  // Nb of substances whose conc. we want to track
  size_t nb_of_cell_types_ =
      13;  // Nb of cell types whose number we want to track
  double sim_time_step = 1;
  double measurement_conc_ops_freq_ = 1;
  double measurement_cells_ops_freq_ = 1;

  // Various
  double M1_MCP1_chemotactic_sensitivity_ = 10.;
  double M1_MCP1_degradation = 0.0;  // 2.08 * pow(10, -4);
  double ECM_saturation = pow(10, -2);
  double IL13_effect_M2_TGFb_secretion = 1.;
  double TGFb_ac_effect_MF_ECM_secretion = 2.;
  double random_movement_probability = 1.;
  double cell_speed = 4. / 60.;                     // um/s
  double fibroblast_migration_speed = 20. / 3600.;  // um/s
  int sub_res = 8;                                  // Substance resolution
  double beta = 0.2;  // influx rate of macrophages into interstitium
  std::string input_file_name = "CellsData_alveolus_new.root";

  // Secretion rates
  double M2_PDGF_secretion =
      2.4 * pow(10, -5) * 3.35 * 10.8;                    // 2.4 * pow(10, -5);
  double M2_MMP_secretion = 3 * pow(10, -4) * 1.55 * 10;  // 3 * pow(10, -4);
  double M2_TIMP_secretion = 6 * pow(10, -5) * 1.64;      // 6 * pow(10, -5);
  double M2_IL13_secretion =
      3.98 * pow(10, -4) * 0.2 * 13.6;  // 3.98 * pow(10, -4)
  double L_IL13_secretion =
      2.12 * pow(10, -7) * 0.3 * 13.6;  // 2.12 * pow(10, -7)
  double M2_TGFb_secretion =
      1.5 * pow(10, -2) * 8.4 * pow(10, -6) *
      2.87;  // 1.5 * pow(10, -2)* 8*pow(10,-12);//1.5 * pow(10, -2);
  double F_TGFb_secretion =
      7.5 * pow(10, -3) * 9.1 * pow(10, -7) * 9.9;  // 7.5 * pow(10, -3);
  double M1_TNFa_secretion =
      1.39 * pow(10, -5) * 15 * 7.35;  // 1.39 * pow(10, -5);
  double AEC2_TNFa_secretion =
      6.9 * pow(10, -6) * 31 * 7.35;  // 6.9 * pow(10, -6);
  double AEC2_MCP1_secretion = 1 * pow(10, -8);
  double AEC2_FGF2_secretion =
      1.23 * pow(10, -8) * 23.8;  // 1.28 * pow(10, -10);
  double AEC2_TGFb_secretion = 5.52 * pow(10, -16);
  double F_ECM_secretion = 3 * pow(10, -3) * 1000;  // 3 * pow(10, -3);
  double MF_ECM_secretion =
      6 * pow(10, -3);  // 6 * pow(10, -3), F_ECM_secretion*2

  // Dissociation constants
  double k_IL13 = 2 * pow(10, -7);
  double k_TGFb_ac = pow(10, -10);
  double k_TGFb_in = pow(10, -10);
  double k_TGFb_in_AEC2_FGF2 = 0.3 * pow(10, -9);
  double k_MCP1 = 5 * pow(10, -9);
  double k_FGF2 = 1.72 * pow(10, -9);
  double k_PDGF = 1.5 * pow(10, -8);
  double k_TNFa = 5 * pow(10, -7);

  // Proliferation rates
  double F_proliferation_IL13_TGFb = 0.00102881;
  double F_proliferation_FGF2 = 57.80258333;
  double F_base_proliferation = 0;
  double AEC2_proliferation = 0.3466;
  double M0_basic_prob_production = 0.025;
  double F_production = 0.000601449;

  // Differentiation rates
  double AEC2_AEC1_differentiation = 0.4753;
  double F_MF_TGFb_ac_differentiation = 0.12;
  double F_MF_PDGF_differentiation = 0.12;
  double M1_M2_differentiation = 9.02 * pow(10, -6);
  double M2_M1_TNFa_differentiation = 5 * pow(10, -3);
  double M0_M1_differentiation = 1.;
  double AEC2_EMT_ECM_differentiation = ECM_saturation / 2.;
  double AEC2_EMT_constant_rate = 1.65 * pow(10, -3);
  double AEC2_F = 0.5;
  double AEC2_MF = 0.5;

  // Apoptosis rates
  double AEC1_apoptosis = 0.4274;
  double AEC2_apoptosis = 0.4274;
  double active_AEC2_apoptosis = 0.0;
  double F_apoptosis = 1.66 * pow(10, -2);
  double M2_apoptosis = 1.5 * pow(10, -2);
  double M1_apoptosis = 2 * pow(10, -2);
  double MF_apoptosis = 1.66 * pow(10, -2);

  // Damage/Repair
  double ROS_death_AEC2 = 0.;  // Increased apoptosis caused by oxidative stress
  double TGFb_ac_death_AEC2 = 1.65 * pow(10, -3);
  int damageType = 0;
  double clusteredDamageSize = 0.5;
  bool setInitialDamageAEC2 = true;  // Decide if there's a damaged zone, first
  double hitProbability =
      27e-3;  // Fraction of cells which are in damaged zone
              // (at the beginning of the simulation), second
  double activationProbability_AEC2 =
      0.25;  // Rate (=probability/time of activation)
  double fibrocytesRepair_AEC2 = pow(10, -3);

  // Depletion rates
  double TIMP_MMP_binding = 4.98 * pow(10, 8);
  double MMP_TIMP_binding = 1.04 * pow(10, 9);
  double MMP_ECM_binding = 2.59 * pow(10, 7);

  // Substances initial concentrations
  double flat_conc = 0.;
  double ECM_flat_conc = 3.26 * pow(10, -3);
  double MCP1_flat_conc = 0.0;
  double PDGF_flat_conc = 0.35 * pow(10, -8);
  double MMP_flat_conc = 0.375 * pow(10, -7);
  double TGFb_in_flat_conc = 2.51 * pow(10, -12);
  double TGFb_ac_flat_conc = 2.51 * pow(10, -12);
  double TIMP_flat_conc = 5.7 * pow(10, -10);
  double TNFa_flat_conc = 2.5 * pow(10, -8);
  double IL13_flat_conc = 3.2 * pow(10, -8);
  double FGF2_flat_conc = 0.5 * pow(10, -12);

  // Substances diffusion constants
  double PDGF_D = pow(10, 8) * 0.0864;     // um^2/day
  double MMP_D = pow(10, 8) * 0.0432;      // um^2/day
  double ECM_D = pow(10, 8) * 0;           // um^2/day
  double MCP1_D = pow(10, 8) * 0.1728;     // um^2/day
  double TIMP_D = pow(10, 8) * 0.0432;     // um^2/day
  double TGFb_in_D = pow(10, 8) * 0.0432;  // um^2/day
  double TGFb_ac_D = pow(10, 8) * 0.0432;  // um^2/day
  double TNFa_D = pow(10, 8) * 0.0129;     // um^2/day
  double IL13_D = pow(10, 8) * 0.0108;     // um^2/day
  double FGF2_D = 65.;                     // um^2/s

  // Substances decay constants
  double PDGF_d = 3.84;    // day^(-1)
  double MMP_d = 4.32;     // day^(-1)
  double ECM_d = 0.37;     // day^(-1)
  double MCP1_d = 1.73;    // day^(-1)
  double TIMP_d = 21.6;    // day^(-1)
  double TGFb_in_d = 11;   // day^(-1)
  double TGFb_ac_d = 499;  // day^(-1)
  double TNFa_d = 55.45;   // day^(-1)
  double IL13_d = 12.47;   // day^(-1)
  double FGF2_d = 1.66;    // day^(-1)

  int n_of_cells = 5;

 protected:
  /// Assign values from config file to variables
  void AssignFromConfig(const std::shared_ptr<cpptoml::table>&) override;

};

}  // namespace bdm

#endif  // SIM_PARAM_H_
