// -----------------------------------------------------------------------------
//
// Copyright (C) Nicolò Cogno.
// All Rights Reserved.
//
// -----------------------------------------------------------------------------

#include "output_utils.h"

namespace bdm {

std::string cellType_to_string(int cell_type) {
  switch (cell_type) {
    case 1:
      return "Alveolar epithelial cells 1";
    case 2:
      return "Alveolar epithelial cells 2";
    case 3:
      return "Endothelial cells";
    case 4:
      return "Fibroblasts";
    case 5:
      return "Alveolar macrophages";
    case 6:
      return "Pericytes";
    case 7:
      return "Interstitial macrophages";
    case 8:
      return "Lymphocytes";
    case 9:
      return "Plasma cells";
    case 10:
      return "Mast cells";
    case 11:
      return "Myofibroblasts";
    case 12:
      return "Monocytes";
    case 13:
      return "Damaged AEC2";
    case 14:
      return "Activated AEC2";
    default:
      return "Unknown cell type";
  }
}

void export_to_CSV(std::string fileName, std::vector<ConcData*> conc,
                   const std::string& outputDir) {
  // Create file
  std::ofstream str(
      Concat(outputDir, "/", std::string(fileName), ".txt").c_str());

  // Create header
  str << "Time,";
  for (size_t i = 0; i < conc.size(); i++) {
    str << conc[i]->sub_name_ << ",";
  }
  str << "\n";

  // Add data
  for (size_t j = 0; j < conc[0]->mean_conc_.size(); j++) {
    str << conc[0]->time_[j] << ",";
    for (uint64_t k = 0; k < conc.size(); k++) {
      str << conc[k]->mean_conc_[j] << ",";
    }
    str << "\n";
  }
  str.close();
}

void export_to_root_concs(std::string fileName, std::vector<ConcData*>* conc,
                          const std::string& outputDir) {
  size_t subs_number = (*conc).size();
  size_t time;
  double conc_subs[subs_number];

  TFile* file =
      new TFile(Concat(outputDir, "/", std::string(fileName), ".root").c_str(),
                "RECREATE");
  TTree* tree =
      new TTree("Concentrations_tree",
                "Root tree containing substances concentration over time");

  tree->Branch("time", &time, "time/i");
  for (size_t j = 0; j < subs_number; j++) {
    tree->Branch((*conc)[j]->sub_name_.c_str(), &conc_subs[j],
                 Concat((*conc)[j]->sub_name_, "/D").c_str());
  }

  for (unsigned long int event = 0; event < (*conc)[0]->mean_conc_.size();
       event++) {
    time = (*conc)[0]->time_[event];
    for (size_t i = 0; i < subs_number; i++) {
      conc_subs[i] = (*conc)[i]->mean_conc_[event];
    }
    tree->Fill();
  }

  // Write/Export tree
  tree->Write();
  file->Close();
}

void plot_conc(std::string plot_fileName, std::vector<ConcData*> conc,
               const std::string& outputDir) {
  for (size_t i = 0; i < conc.size(); i++) {
    TCanvas* c = new TCanvas();
    c->SetGrid();

    // Plot
    uint64_t n = conc[0]->time_.size();

    TGraph* gr =
        new TGraph(n, conc[i]->time_.data(), conc[i]->mean_conc_.data());
    gr->SetTitle(Concat(conc[i]->sub_name_, ";t [s];#rho [g/mL]").c_str());
    gr->SetLineWidth(1);
    gr->GetXaxis()->SetMaxDigits(3);
    gr->GetYaxis()->SetMaxDigits(3);
    gr->GetXaxis()->SetTitleOffset(3);
    gr->GetYaxis()->SetTitleOffset(4.5);

    if (conc[i]->sub_name_ == "PDGF") {
      gr->SetMaximum(1 * pow(10, -8));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "MMP") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "ECM") {
      gr->SetMaximum(6 * pow(10, -3));
      gr->SetMinimum(2 * pow(10, -3));
    }
    if (conc[i]->sub_name_ == "MCP-1") {
      gr->SetMaximum(4 * pow(10, -9));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TIMP") {
      gr->SetMaximum(10 * pow(10, -10));
      gr->SetMinimum(0 * pow(10, -10));
    }
    if (conc[i]->sub_name_ == "TNFa") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TGFb inactive") {
      gr->SetMaximum(1 * pow(10, -11));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TGFb active") {
      gr->SetMaximum(1 * pow(10, -11));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "FGF-2") {
      gr->SetMaximum(1 * pow(10, -12));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "IL-13") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }

    gr->SetMarkerColor(4);
    gr->SetMarkerSize(0.5);
    gr->SetMarkerStyle(20);

    gr->Draw("apl");

    c->Update();
    c->Modified();
    c->SaveAs(Concat(outputDir, "/", std::string(plot_fileName), "_",
                     conc[i]->sub_name_, ".png")
                  .c_str());

    delete gr;
    delete c;
  }
}

void plot_conc_all_in_one(std::string plot_fileName,
                          std::vector<ConcData*> conc,
                          const std::string& outputDir) {

  TCanvas* c1 = new TCanvas("c1", "multipads", 2560, 1600);

  c1->SetGrid();
  c1->Divide(4, 3, 0,
             0);

  gStyle->SetTitleX(0.6);
  gStyle->SetTitleY(1.01);
  gStyle->SetTitleFont(21, "");

  int colors[] = {1, 2, 3, 4, 6, 7, 8, 9, 30, 38, 39, 46, 49, 28, 5};

  for (size_t i = 0; i < conc.size(); i++) {
    uint64_t n = conc[0]->time_.size();

    TGraph* gr =
        new TGraph(n, conc[i]->time_.data(), conc[i]->mean_conc_.data());
    gr->SetTitle((conc[i]->sub_name_).c_str());

    gr->GetXaxis()->SetTitleFont(43);
    gr->GetYaxis()->SetTitleFont(43);
    gr->GetXaxis()->SetLabelFont(43);
    gr->GetYaxis()->SetLabelFont(43);
    gr->GetXaxis()->SetLabelSize(20);
    gr->GetYaxis()->SetLabelSize(20);
    gr->GetXaxis()->SetTitleSize(22);
    gr->GetYaxis()->SetTitleSize(22);
    gr->GetXaxis()->SetTitleOffset(3);
    gr->GetYaxis()->SetTitleOffset(4.5);

    gPad->Update();

    gr->SetLineColor(colors[i]);
    gr->SetLineWidth(1);

    if (i == 0 || i == 4 || i == 8) {
      gr->GetYaxis()->SetTitle("#rho [g/mL]");
    }

    gr->GetXaxis()->SetTitle("t [s]");
    gr->GetXaxis()->SetMaxDigits(3);
    gr->GetYaxis()->SetMaxDigits(3);


    // gr->SetFillColor(kBlue);
    c1->cd(i + 1);
    gPad->Update();

    if (conc[i]->sub_name_ == "PDGF") {
      gr->SetMaximum(1 * pow(10, -8));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "MMP") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "ECM") {
      gr->SetMaximum(6 * pow(10, -3));
      gr->SetMinimum(2 * pow(10, -3));
    }
    if (conc[i]->sub_name_ == "MCP-1") {
      gr->SetMaximum(4 * pow(10, -9));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TIMP") {
      gr->SetMaximum(10 * pow(10, -10));
      gr->SetMinimum(0 * pow(10, -10));
    }
    if (conc[i]->sub_name_ == "TNFa") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TGFb inactive") {
      gr->SetMaximum(1 * pow(10, -11));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "TGFb active") {
      gr->SetMaximum(1 * pow(10, -11));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "FGF-2") {
      gr->SetMaximum(1 * pow(10, -12));
      gr->SetMinimum(0.0);
    }
    if (conc[i]->sub_name_ == "IL-13") {
      gr->SetMaximum(1 * pow(10, -7));
      gr->SetMinimum(0.0);
    }

    gPad->SetLeftMargin(0.13);
    gPad->SetRightMargin(0.09);
    gPad->SetTopMargin(0.05);
    gPad->SetBottomMargin(0.1);
    gr->SetMarkerColor(colors[i]);
    gr->SetMarkerSize(0.5);
    gr->SetMarkerStyle(20);

    TF1* fitLine = new TF1("Linear fit", "pol1", 0, 3600);
    fitLine->SetLineColor(kRed);
    fitLine->SetLineStyle(2);
    fitLine->SetParName(0, "q");
    fitLine->SetParName(1, "m");

    gr->Fit(fitLine, "Q");

    gr->Draw("apl");

    double y_ss = 0.0;

    if (conc[i]->sub_name_ == "PDGF") {
      y_ss = 0.35 * pow(10, -8);
    }
    if (conc[i]->sub_name_ == "MMP") {
      y_ss = 0.375 * pow(10, -7);
    }
    if (conc[i]->sub_name_ == "ECM") {
      y_ss = 3.26 * pow(10, -3);
    }
    if (conc[i]->sub_name_ == "MCP-1") {
      y_ss = 0.0;
    }
    if (conc[i]->sub_name_ == "TIMP") {
      y_ss = 5.7 * pow(10, -10);
    }
    if (conc[i]->sub_name_ == "TNFa") {
      y_ss = 2.5 * pow(10, -8);
    }
    if (conc[i]->sub_name_ == "TGFb inactive") {
      y_ss = 2.51 * pow(10, -12);
    }
    if (conc[i]->sub_name_ == "TGFb active") {
      y_ss = 2.51 * pow(10, -12);
    }
    if (conc[i]->sub_name_ == "FGF-2") {
      y_ss = 0.5 * pow(10, -12);
    }
    if (conc[i]->sub_name_ == "IL-13") {
      y_ss = 3.2 * pow(10, -8);
    }
    TLine* l = new TLine(gr->GetXaxis()->GetXmin(), y_ss,
                         gr->GetXaxis()->GetXmax(), y_ss);
    l->SetLineColor(kBlue);
    l->SetLineStyle(10);
    l->Draw();

    // Legend

    auto legend = new TLegend(0.25, 0.8, 0.91, 0.95);

    legend->AddEntry(gr, "Simulation data", "lp");
    legend->AddEntry(l, "Ss from literature", "l");
    Char_t q[20], m[20];
    sprintf(q, "%.3g", fitLine->GetParameter(0));
    sprintf(m, "%.3g", fitLine->GetParameter(1));
    legend->AddEntry(fitLine, Concat("Lin fit: q = ", q, ", m = ", m).c_str(),
                     "l");
    legend->SetTextFont(
        43);
    legend->SetTextSize(18);
    legend->Draw();

    c1->Update();
  }

  c1->SaveAs(
      Concat(outputDir, "/", std::string(plot_fileName), "_", "All", ".png")
          .c_str());
}

void plot_cells(std::string plot_fileName, std::vector<std::vector<int>> cells,
                size_t time_steps, size_t cell_types, size_t op_freq,
                const std::string& outputDir) {
  std::vector<int> time_;

  for (size_t i = 0; i < time_steps; i++) {
    time_.push_back(i * op_freq);
  }

  int colors[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 30, 38, 39, 49, 46, 28};

  // setup plot
  TCanvas* c = new TCanvas("c", "c", 1200, 900);

  c->SetGrid();
  TMultiGraph* mg = new TMultiGraph();

  mg->SetTitle(";t [s];Cells [#]");

  // Plot
  uint64_t n = cells[0].size();

  TGraph* gr[cell_types];
  auto legend = new TLegend(0.53, 0.6, 0.98, 0.8);
  legend->SetNColumns(2);

  for (size_t i = 0; i < cells.size(); i++) {
    gr[i] = new TGraph(n, time_.data(), cells[i].data());
    gr[i]->SetTitle(
        Concat(cellType_to_string(i + 1), ";t [s];Cells [#]").c_str());
    gr[i]->SetLineColor(colors[i]);
    gr[i]->SetLineWidth(2);
    mg->Add(gr[i]);
    legend->AddEntry(gr[i], cellType_to_string(i + 1).c_str(), "l");
  }

  mg->GetYaxis()->SetNdivisions(35);
  mg->GetXaxis()->SetLabelSize(0.03);
  mg->GetYaxis()->SetLabelSize(0.03);

  mg->Draw("al");

  legend->Draw();

  gPad->Update();
  gPad->Modified();

  c->SaveAs(Concat(outputDir, "/", std::string(plot_fileName), ".png").c_str());
}

void export_to_CSV_cells(std::string fileName,
                         std::vector<std::vector<int>> cells, size_t time_steps,
                         size_t cell_types, size_t op_freq,
                         const std::string& outputDir) {
  std::ios_base::sync_with_stdio(false);
  std::vector<int> time_;
  size_t cell_number = cells.size();

  for (size_t i = 0; i < time_steps; i++) {
    time_.push_back(i * op_freq);
  }

  std::ofstream str(
      Concat(outputDir, "/", std::string(fileName), ".txt").c_str());
  str << "Time, ";
  for (size_t j = 0; j < cell_number; j++) {
    str << cellType_to_string(j + 1) << ", ";
  }

  str << "\n";

  for (uint64_t k = 0; k < cells[0].size(); k++) {
    str << time_[k] << ",";

    for (size_t i = 0; i < cell_number; i++) {
      str << cells[i][k] << ",";
    }
    str << "\n";
  }

  str.close();
}

void export_to_root_cells(std::string fileName,
                          std::vector<std::vector<double>> cells,
                          size_t time_steps, size_t cell_types, size_t op_freq,
                          const std::string& outputDir, int nb_of_alveoli) {
  size_t time, number_cell_type[cell_types];
  std::vector<int> time_;
  size_t cell_number = cells.size();

  auto* sim = Simulation::GetActive();
  auto* param = sim->GetParam();

  for (size_t i = 0; i < time_steps; i++) {
    time_.push_back(i * op_freq * param->simulation_time_step);
  }

  TFile* file =
      new TFile(Concat(outputDir, "/", std::string(fileName), ".root").c_str(),
                "RECREATE");
  TTree* tree =
      new TTree("Cell_number_tree",
                "Root tree containing cell numbers (per type) over time");

  tree->Branch("time", &time, "time/i");
  for (size_t j = 0; j < cell_number; j++) {
    tree->Branch(cellType_to_string(j + 1).c_str(), &number_cell_type[j],
                 Concat(cellType_to_string(j + 1), "/i").c_str());
  }

  for (unsigned long int event = 0; event < cells[0].size(); event++) {
    time = time_[event];
    for (size_t i = 0; i < cell_number; i++) {
      number_cell_type[i] = cells[i][event];
    }

    tree->Fill();
  }

  TTree* tree1 = new TTree("Alveoli_number_tree", "Number of alveoli");
  Int_t alveoli_;
  tree1->Branch("alveoli_", &alveoli_, "alveoli_/I");
  alveoli_ = nb_of_alveoli;
  tree1->Fill();

  // Write/Export tree
  tree->Write();
  tree1->Write();
  file->Close();
}

void export_to_root_local_concs(std::string fileName,
                                std::vector<std::vector<double>> local_concs,
                                size_t time_steps, size_t nb_of_subs,
                                size_t op_freq, const std::string& outputDir) {
  size_t time;
  double local_conc_sub[nb_of_subs];
  std::vector<int> time_;
  size_t subs_number = local_concs.size();
  ConcData dummyVar;  // Just because we need the "enum_to_string" function

  auto* sim = Simulation::GetActive();
  auto* param = sim->GetParam();

  for (size_t i = 0; i < time_steps; i++) {
    time_.push_back(i * op_freq * param->simulation_time_step);
  }

  TFile* file =
      new TFile(Concat(outputDir, "/", std::string(fileName), ".root").c_str(),
                "RECREATE");
  TTree* tree = new TTree("Local_conc_tree",
                          "Root tree containing local concs of subs over time");

  tree->Branch("time", &time, "time/i");
  for (size_t j = 0; j < subs_number; j++) {
    dummyVar.enum_to_string(static_cast<Substances>(j));
    tree->Branch(dummyVar.sub_name_.c_str(), &local_conc_sub[j],
                 Concat(dummyVar.sub_name_.c_str(), "/D").c_str());
  }

  for (unsigned long int event = 0; event < local_concs[0].size(); event++) {
    time = time_[event];
    for (size_t i = 0; i < subs_number; i++) {
      local_conc_sub[i] = local_concs[i][event];
    }

    tree->Fill();
  }

  // Write/Export tree
  tree->Write();
  file->Close();
}
